# encoding=UTF-8

# Copyright © 2010-2016 Jakub Wilk <jwilk@jwilk.net>
#
# This file is part of python-djvulibre.
#
# python-djvulibre is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# python-djvulibre is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.

import codecs
import contextlib
import distutils.spawn
import locale
import os
import re
import sys

from nose import SkipTest

import nose.tools

from nose.tools import (
    assert_true,
    assert_false,
    assert_equal,
    assert_not_equal,
)

def noseimport(vmaj, vmin, name=None):
    def wrapper(f):
        if f.__module__ == 'unittest.case':
            return f
        if sys.version_info >= (vmaj, vmin):
            return getattr(nose.tools, name or f.__name__)
        return f
    return wrapper

@noseimport(2, 7)
def assert_in(x, y):
    assert_true(
        x in y,
        msg='{0!r} not found in {1!r}'.format(x, y)
    )

@noseimport(2, 7)
def assert_is(x, y):
    assert_true(
        x is y,
        msg='{0!r} is not {1!r}'.format(x, y)
    )

@noseimport(2, 7)
def assert_is_instance(obj, cls):
    assert_true(
        isinstance(obj, cls),
        msg='{0!r} is not an instance of {1!r}'.format(obj, cls)
    )

@noseimport(2, 7)
def assert_less(x, y):
    assert_true(
        x < y,
        msg='{0!r} not less than {1!r}'.format(x, y)
    )

@noseimport(2, 7)
def assert_list_equal(x, y):
    assert_is_instance(x, list)
    assert_is_instance(y, list)
    return assert_equal(x, y)

@noseimport(2, 7)
def assert_multi_line_equal(x, y):
    return assert_equal(x, y)
if sys.version_info >= (2, 7):
    type(assert_multi_line_equal.__self__).maxDiff = None

@noseimport(2, 7)
def assert_not_in(x, y):
    assert_true(
        x not in y,
        msg='{0!r} unexpectedly found in {1!r}'.format(x, y)
    )

@noseimport(2, 7)
class assert_raises(object):
    def __init__(self, exc_type):
        self._exc_type = exc_type
        self.exception = None
    def __enter__(self):
        return self
    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is None:
            assert_true(False, '{0} not raised'.format(self._exc_type.__name__))
        if not issubclass(exc_type, self._exc_type):
            return False
        if isinstance(exc_value, exc_type):
            pass
            # This branch is not always taken in Python 2.6:
            # https://bugs.python.org/issue7853
        elif isinstance(exc_value, tuple):
            exc_value = exc_type(*exc_value)
        else:
            exc_value = exc_type(exc_value)
        self.exception = exc_value
        return True

@noseimport(2, 7, 'assert_raises_regexp')
@noseimport(3, 2)
@contextlib.contextmanager
def assert_raises_regex(exc_type, regex):
    with assert_raises(exc_type) as ecm:
        yield
    assert_regex(str(ecm.exception), regex)

@noseimport(2, 7, 'assert_regexp_matches')
@noseimport(3, 2)
def assert_regex(text, regex):
    if isinstance(regex, basestring):
        regex = re.compile(regex)
    if not regex.search(text):
        message = "Regex didn't match: {0!r} not found in {1!r}".format(regex.pattern, text)
        assert_true(False, msg=message)

@contextlib.contextmanager
def assert_raises_str(exc_type, s):
    with assert_raises(exc_type) as ecm:
        yield
    assert_equal(str(ecm.exception), s)

try:
    locale.LC_MESSAGES
except AttributeError:
    # A non-POSIX system.
    locale.LC_MESSAGES = locale.LC_ALL

locale_encoding = locale.getpreferredencoding()
if codecs.lookup(locale_encoding) == codecs.lookup('US-ASCII'):
    locale_encoding = 'UTF-8'

py3k = sys.version_info >= (3, 0)

if py3k:
    def u(s):
        return s
else:
    def u(s):
        return s.decode('UTF-8')

if py3k:
    def b(s):
        return s.encode('UTF-8')
else:
    def b(s):
        return s

if py3k:
    long = int
else:
    long = long

if py3k:
    def cmp(x, y):
        if x == y:
            return 0
        if x < y:
            return -1
        if x > y:
            return 1
        assert False
else:
    cmp = cmp

if py3k:
    from io import StringIO
else:
    from io import BytesIO as StringIO

if py3k:
    unicode = str
else:
    unicode = unicode

if py3k:
    maxsize = sys.maxsize
else:
    maxsize = sys.maxint

@contextlib.contextmanager
def interim(obj, **override):
    copy = dict((key, getattr(obj, key)) for key in override)
    for key, value in override.items():
        setattr(obj, key, value)
    try:
        yield
    finally:
        for key, value in copy.items():
            setattr(obj, key, value)

@contextlib.contextmanager
def interim_locale(**kwargs):
    old_locale = locale.setlocale(locale.LC_ALL)
    try:
        for category, value in kwargs.items():
            category = getattr(locale, category)
            try:
                locale.setlocale(category, value)
            except locale.Error as exc:
                raise SkipTest(exc)
        yield
    finally:
        locale.setlocale(locale.LC_ALL, old_locale)

def assert_repr(self, expected):
    return assert_equal(repr(self), expected)

def skip_unless_c_messages():
    if locale.setlocale(locale.LC_MESSAGES) != 'C':
        raise SkipTest('you need to run this test with LC_MESSAGES=C')
    if os.getenv('LANGUAGE', '') != '':
        raise SkipTest('you need to run this test with LANGUAGE unset')

def skip_unless_translation_exists(lang):
    messages = {}
    langs = ['C', lang]
    for lang in langs:
        with interim_locale(LC_ALL=lang):
            try:
                open(__file__ + '/')
            except EnvironmentError as exc:
                messages[lang] = str(exc)
    messages = set(messages.values())
    assert 1 <= len(messages) <= 2
    if len(messages) == 1:
        raise SkipTest('libc translation not found: ' + lang)

def skip_unless_command_exists(command):
    if distutils.spawn.find_executable(command):
        return
    raise SkipTest('command not found: ' + command)

def wildcard_import(mod):
    ns = {}
    exec('from {mod} import *'.format(mod=mod), {}, ns)
    return ns

__all__ = [
    # Python 2/3 compat:
    'StringIO',
    'b',
    'cmp',
    'long',
    'maxsize',
    'py3k',
    'u',
    'unicode',
    # nose
    'SkipTest',
    'assert_equal',
    'assert_false',
    'assert_in',
    'assert_is',
    'assert_is_instance',
    'assert_less',
    'assert_list_equal',
    'assert_multi_line_equal',
    'assert_not_equal',
    'assert_not_in',
    'assert_raises',
    'assert_raises_regex',
    'assert_regex',
    'assert_true',
    # misc
    'assert_raises_str',
    'assert_repr',
    'interim',
    'interim_locale',
    'locale_encoding',
    'skip_unless_c_messages',
    'skip_unless_command_exists',
    'skip_unless_translation_exists',
    'wildcard_import',
]

# vim:ts=4 sts=4 sw=4 et



# encoding=UTF-8

# Copyright © 2011-2016 Jakub Wilk <jwilk@jwilk.net>
#
# This file is part of python-djvulibre.
#
# python-djvulibre is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# python-djvulibre is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.

'''
ease finding DjVuLibre DLLs in non-standard locations
'''

import os

if os.name != 'nt':
    raise ImportError('This module is for Windows only')

import ctypes

try:
    # Python 3.X
    import winreg
    unicode = str
except ImportError:
    # Python 2.X
    import _winreg as winreg

def _get(key, subkey):
    unicode = type(b''.decode())
    with winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE) as registry:
        with winreg.OpenKey(registry, key) as regkey:
            value, tp = winreg.QueryValueEx(regkey, subkey)
            del tp
            if not isinstance(value, unicode):
                raise TypeError
            return value

_djvulibre_key = r'Software\Microsoft\Windows\CurrentVersion\Uninstall\DjVuLibre+DjView'

def guess_dll_path():
    try:
        path = _get(_djvulibre_key, 'UninstallString')
    except (TypeError, WindowsError):
        return
    path = os.path.dirname(path)
    if os.path.isfile(os.path.join(path, 'libdjvulibre.dll')):
        return path

def _guess_dll_version():
    try:
        version = _get(_djvulibre_key, 'DisplayVersion')
    except (TypeError, WindowsError):
        return
    return version.split('+')[0]

def set_dll_search_path(path=None):
    unicode = type(b''.decode())
    if path is None:
        path = guess_dll_path()
    if path is None:
        return
    if not isinstance(path, unicode):
        raise TypeError
    ctypes.windll.kernel32.SetDllDirectoryW(path)
    return path

__all__ = [
    'guess_dll_path',
    'set_dll_search_path'
]

# vim:ts=4 sts=4 sw=4 et



# Copyright © 2008-2015 Jakub Wilk <jwilk@jwilk.net>
#
# This file is part of python-djvulibre.
#
# python-djvulibre is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# python-djvulibre is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.

include 'config.pxi'

# C library:

from libc.stdlib cimport free
from libc.string cimport strlen

# Python memory handling:

from cpython.mem cimport PyMem_Malloc as py_malloc
from cpython.mem cimport PyMem_Free as py_free

# Python numbers:

from cpython cimport (
    PyInt_Check as is_short_int,
    PyLong_Check as is_long_int,
)
cdef int is_int(object o):
    return is_short_int(o) or is_long_int(o)

from cpython cimport (
    PyNumber_Check as is_number,
    PyFloat_Check as is_float,
)

IF PY3K:
    from cpython cimport PyNumber_Long as int
ELSE:
    from cpython cimport PyNumber_Int as int
    from cpython cimport PyNumber_Long as long

# Python strings:

from cpython cimport (
    PyUnicode_Check as is_unicode,
    PyString_Check as is_string,
    PyBytes_Check as is_bytes,
)

from cpython cimport (
    PyUnicode_AsUTF8String as encode_utf8,
    PyUnicode_DecodeUTF8 as decode_utf8_ex,
    PyBytes_AsStringAndSize as bytes_to_charp,
    PyBytes_FromStringAndSize as charp_to_bytes,
)
IF PY3K:
    cdef extern from 'Python.h':
        object charp_to_string 'PyUnicode_FromString'(char *v)
ELSE:
    from cpython cimport PyString_FromString as charp_to_string

cdef object decode_utf8(const char *s):
    return decode_utf8_ex(s, strlen(s), NULL)

cdef extern from 'Python.h':
    int buffer_to_writable_memory 'PyObject_AsWriteBuffer'(object, void **, Py_ssize_t *)

# Python booleans:

from cpython cimport PyBool_FromLong as bool

# Python pointer->integer conversion:

from cpython cimport PyLong_FromVoidPtr as voidp_to_int

# Python files:

from libc.stdio cimport FILE

# Python lists:

from cpython cimport PyList_Append as list_append

# Python rich comparison:

from cpython cimport PyObject_RichCompare as richcmp

# Python slices:

cdef extern from 'Python.h':
    int is_slice 'PySlice_Check'(object)

# Python threads:

from cpython cimport (
    PyThread_type_lock as Lock,
    PyThread_allocate_lock as allocate_lock,
    PyThread_free_lock as free_lock,
    PyThread_acquire_lock as acquire_lock,
    PyThread_release_lock as release_lock,
    WAIT_LOCK,
    NOWAIT_LOCK,
)

# Python type checks:

cdef extern from 'object.h':
    ctypedef struct PyTypeObject:
        const char *tp_name

from cpython cimport PyObject
from cpython cimport PyObject_TypeCheck as _typecheck

cdef object type(object o):
    return <object>((<PyObject*>o).ob_type)

IF PY3K:
    cdef object get_type_name(object type):
        return decode_utf8((<PyTypeObject*>type).tp_name)
ELSE:
    cdef const char* get_type_name(object type):
        return (<PyTypeObject*>type).tp_name

cdef int typecheck(object o, object type):
    return _typecheck(o, <PyTypeObject*> type)

# Python exceptions:

cdef void raise_instantiation_error(object cls) except *:
    raise TypeError, 'cannot create \'{tp}\' instances'.format(tp=get_type_name(cls))

cdef extern from 'pyerrors.h':
    ctypedef class __builtin__.Exception [object PyBaseExceptionObject]:
        pass

# vim:ts=4 sts=4 sw=4 et ft=pyrex



# encoding=UTF-8

# Copyright © 2015 Jakub Wilk <jwilk@jwilk.net>
#
# This file is part of python-djvulibre.
#
# python-djvulibre is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# python-djvulibre is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.

import sys

if sys.version_info < (2, 6):
    raise RuntimeError('Python >= 2.6 is required')

# vim:ts=4 sts=4 sts=4 sw=4 et



# encoding=UTF-8

# Copyright © 2008-2015 Jakub Wilk <jwilk@jwilk.net>
#
# This file is part of python-djvulibre.
#
# python-djvulibre is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# python-djvulibre is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.

'''DjVuLibre bindings: various constants.'''

import djvu.sexpr

EMPTY_LIST = djvu.sexpr.Expression([])
EMPTY_OUTLINE = djvu.sexpr.Expression([djvu.sexpr.Symbol('bookmarks')])

METADATA_BIBTEX_KEYS = frozenset(djvu.sexpr.Symbol(x) for x in '''
address
annote
author
booktitle
chapter
crossref
edition
editor
howpublished
institution
journal
key
month
note
number
organization
pages
publisher
school
series
title
type
volume
year'''.split())
# Retrieved from <https://www.ctan.org/pkg/bibtex>

METADATA_PDFINFO_KEYS = frozenset(djvu.sexpr.Symbol(x) for x in '''
Author
CreationDate
Creator
Keywords
ModDate
Producer
Subject
Title
Trapped'''.split())
# Retrieved from the PDF specification

METADATA_KEYS = METADATA_BIBTEX_KEYS | METADATA_PDFINFO_KEYS

class TextZoneType(djvu.sexpr.Symbol):

    '''
    A type of a text zone. You can compare text zone types with the < operator.

    To create objects of this class, use the get_text_zone_type() function.
    '''

    __cache = {}

    @classmethod
    def from_symbol(cls, symbol):
        return cls.__cache[symbol]

    def __new__(cls, value, rank):
        self = djvu.sexpr.Symbol.__new__(cls, value)
        TextZoneType.__cache[self] = self
        return self

    def __init__(self, value, rank):
        self.__rank = rank

    def __lt__(self, other):
        if not isinstance(other, TextZoneType):
            raise TypeError('cannot compare text zone type with other object')
        return self.__rank < other.__rank

    def __le__(self, other):
        if not isinstance(other, TextZoneType):
            raise TypeError('cannot compare text zone type with other object')
        return self.__rank <= other.__rank

    def __gt__(self, other):
        if not isinstance(other, TextZoneType):
            raise TypeError('cannot compare text zone type with other object')
        return self.__rank > other.__rank

    def __ge__(self, other):
        if not isinstance(other, TextZoneType):
            raise TypeError('cannot compare text zone type with other object')
        return self.__rank >= other.__rank

    def __repr__(self):
        return '<{mod}.{cls}: {name}>'.format(
            mod=self.__module__,
            cls=type(self).__name__,
            name=self
        )

TEXT_ZONE_PAGE = TextZoneType('page', 7)
TEXT_ZONE_COLUMN = TextZoneType('column', 6)
TEXT_ZONE_REGION = TextZoneType('region', 5)
TEXT_ZONE_PARAGRAPH = TextZoneType('para', 4)
TEXT_ZONE_LINE = TextZoneType('line', 3)
TEXT_ZONE_WORD = TextZoneType('word', 2)
TEXT_ZONE_CHARACTER = TextZoneType('char', 1)

def get_text_zone_type(symbol):
    return TextZoneType.from_symbol(symbol)

TEXT_ZONE_SEPARATORS = {
    TEXT_ZONE_PAGE: '\f',  # Form Feed (FF)
    TEXT_ZONE_COLUMN: '\v',  # Vertical tab (VT, LINE TABULATION)
    TEXT_ZONE_REGION: '\035',  # Group Separator (GS, INFORMATION SEPARATOR THREE)
    TEXT_ZONE_PARAGRAPH: '\037',  # Unit Separator (US, INFORMATION SEPARATOR ONE)
    TEXT_ZONE_LINE: '\n',  # Line Feed (LF)
    TEXT_ZONE_WORD: ' ',  # space
    TEXT_ZONE_CHARACTER: ''
}

# 8.3.4.2 Maparea (overprinted annotations)
ANNOTATION_MAPAREA = djvu.sexpr.Symbol('maparea')

# 8.3.4.2 Maparea (overprinted annotations):
MAPAREA_SHAPE_RECTANGLE = djvu.sexpr.Symbol('rect')
MAPAREA_SHAPE_OVAL = djvu.sexpr.Symbol('oval')
MAPAREA_SHAPE_POLYGON = djvu.sexpr.Symbol('poly')
MAPAREA_SHAPE_LINE = djvu.sexpr.Symbol('line')
MAPAREA_SHAPE_TEXT = djvu.sexpr.Symbol('text')

MAPAREA_URI = MAPAREA_URL = djvu.sexpr.Symbol('url')

# 8.3.4.2.3.1.1 Border type:
MAPAREA_BORDER_NONE = djvu.sexpr.Symbol('none')
MAPAREA_BORDER_XOR = djvu.sexpr.Symbol('xor')
MAPAREA_BORDER_SOLID_COLOR = djvu.sexpr.Symbol('border')

# 8.3.4.2.3.1.1 Border type:
MAPAREA_BORDER_SHADOW_IN = djvu.sexpr.Symbol('shadow_in')
MAPAREA_BORDER_SHADOW_OUT = djvu.sexpr.Symbol('shadow_out')
MAPAREA_BORDER_ETCHED_IN = djvu.sexpr.Symbol('shadow_ein')
MAPAREA_BORDER_ETCHED_OUT = djvu.sexpr.Symbol('shadow_eout')
MAPAREA_SHADOW_BORDERS = (MAPAREA_BORDER_SHADOW_IN, MAPAREA_BORDER_SHADOW_OUT, MAPAREA_BORDER_ETCHED_IN, MAPAREA_BORDER_ETCHED_OUT)
MAPAREA_SHADOW_BORDER_MIN_WIDTH = 1
MAPAREA_SHADOW_BORDER_MAX_WIDTH = 32

# 8.3.4.2.3.1.2 Border always visible
MAPAREA_BORDER_ALWAYS_VISIBLE = djvu.sexpr.Symbol('border_avis')

# 8.3.4.2.3.1.3 Highlight color and opacity:
MAPAREA_HIGHLIGHT_COLOR = djvu.sexpr.Symbol('hilite')
MAPAREA_OPACITY = djvu.sexpr.Symbol('opacity')
MAPAREA_OPACITY_MIN = 0
MAPAREA_OPACITY_DEFAULT = 50
MAPAREA_OPACITY_MAX = 100

# 8.3.4.2.3.1.4 Line and Text parameters:
MAPAREA_ARROW = djvu.sexpr.Symbol('arrow')
MAPAREA_LINE_WIDTH = djvu.sexpr.Symbol('width')
MAPAREA_LINE_COLOR = djvu.sexpr.Symbol('lineclr')
MAPAREA_LINE_MIN_WIDTH = 1
MAPAREA_LINE_COLOR_DEFAULT = '#000000'

# 8.3.4.2.3.1.4 Line and Text parameters:
MAPAREA_BACKGROUND_COLOR = djvu.sexpr.Symbol('backclr')
MAPAREA_TEXT_COLOR = djvu.sexpr.Symbol('textclr')
MAPAREA_PUSHPIN = djvu.sexpr.Symbol('pushpin')
MAPAREA_TEXT_COLOR_DEFAULT = '#000000'

# 8.3.4.1 Initial Document View :
ANNOTATION_BACKGROUND = djvu.sexpr.Symbol('background')  # 8.3.4.1.1 Background Color
ANNOTATION_ZOOM = djvu.sexpr.Symbol('zoom')  # 8.3.4.1.2 Initial Zoom
ANNOTATION_MODE = djvu.sexpr.Symbol('mode')  # 8.3.4.1.3 Initial Display level
ANNOTATION_ALIGN = djvu.sexpr.Symbol('align')  # 8.3.4.1.4 Alignment

# djvuchanges.txt, sections "Metadata Annotations" and "Document Annotations and Metadata":
ANNOTATION_METADATA = djvu.sexpr.Symbol('metadata')

# 8.3.4.3 Printed headers and footers:
ANNOTATION_PRINTED_HEADER = djvu.sexpr.Symbol('phead')
ANNOTATION_PRINTED_FOOTER = djvu.sexpr.Symbol('pfoot')
PRINTER_HEADER_ALIGN_LEFT = PRINTED_FOOTER_ALIGN_LEFT = djvu.sexpr.Symbol('left')
PRINTER_HEADER_ALIGN_CENTER = PRINTED_FOOTER_ALIGN_CENTER = djvu.sexpr.Symbol('center')
PRINTER_HEADER_ALIGN_RIGHT = PRINTED_FOOTER_ALIGN_RIGHT = djvu.sexpr.Symbol('right')

__all__ = [
    'ANNOTATION_ALIGN',
    'ANNOTATION_BACKGROUND',
    'ANNOTATION_MAPAREA',
    'ANNOTATION_METADATA',
    'ANNOTATION_MODE',
    'ANNOTATION_PRINTED_FOOTER',
    'ANNOTATION_PRINTED_HEADER',
    'ANNOTATION_ZOOM',
    'EMPTY_LIST',
    'EMPTY_OUTLINE',
    'MAPAREA_ARROW',
    'MAPAREA_BACKGROUND_COLOR',
    'MAPAREA_BORDER_ALWAYS_VISIBLE',
    'MAPAREA_BORDER_ETCHED_IN',
    'MAPAREA_BORDER_ETCHED_OUT',
    'MAPAREA_BORDER_NONE',
    'MAPAREA_BORDER_SHADOW_IN',
    'MAPAREA_BORDER_SHADOW_OUT',
    'MAPAREA_BORDER_SOLID_COLOR',
    'MAPAREA_BORDER_XOR',
    'MAPAREA_HIGHLIGHT_COLOR',
    'MAPAREA_LINE_COLOR',
    'MAPAREA_LINE_COLOR_DEFAULT',
    'MAPAREA_LINE_MIN_WIDTH',
    'MAPAREA_LINE_WIDTH',
    'MAPAREA_OPACITY',
    'MAPAREA_OPACITY_DEFAULT',
    'MAPAREA_OPACITY_MAX',
    'MAPAREA_OPACITY_MIN',
    'MAPAREA_PUSHPIN',
    'MAPAREA_SHADOW_BORDERS',
    'MAPAREA_SHADOW_BORDER_MAX_WIDTH',
    'MAPAREA_SHADOW_BORDER_MIN_WIDTH',
    'MAPAREA_SHAPE_LINE',
    'MAPAREA_SHAPE_OVAL',
    'MAPAREA_SHAPE_POLYGON',
    'MAPAREA_SHAPE_RECTANGLE',
    'MAPAREA_SHAPE_TEXT',
    'MAPAREA_TEXT_COLOR',
    'MAPAREA_TEXT_COLOR_DEFAULT',
    'MAPAREA_URI',
    'MAPAREA_URL',
    'METADATA_BIBTEX_KEYS',
    'METADATA_KEYS',
    'METADATA_PDFINFO_KEYS',
    'PRINTED_FOOTER_ALIGN_CENTER',
    'PRINTED_FOOTER_ALIGN_LEFT',
    'PRINTED_FOOTER_ALIGN_RIGHT',
    'PRINTER_HEADER_ALIGN_CENTER',
    'PRINTER_HEADER_ALIGN_LEFT',
    'PRINTER_HEADER_ALIGN_RIGHT',
    'TEXT_ZONE_CHARACTER',
    'TEXT_ZONE_COLUMN',
    'TEXT_ZONE_LINE',
    'TEXT_ZONE_PAGE',
    'TEXT_ZONE_PARAGRAPH',
    'TEXT_ZONE_REGION',
    'TEXT_ZONE_SEPARATORS',
    'TEXT_ZONE_WORD',
    'TextZoneType',
    'get_text_zone_type'
]

# vim:ts=4 sts=4 sw=4 et



# Copyright © 2007-2015 Jakub Wilk <jwilk@jwilk.net>
#
# This file is part of python-djvulibre.
#
# python-djvulibre is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# python-djvulibre is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.

cdef extern from 'stdio.h':
    ctypedef struct FILE

from djvu.sexpr cimport cexpr_t, _WrappedCExpr
from djvu.sexpr cimport public_c2py as cexpr2py
from djvu.sexpr cimport public_py2c as py2cexpr

cdef extern from 'libdjvu/ddjvuapi.h':
    struct ddjvu_context_s
    union ddjvu_message_s
    struct ddjvu_job_s
    struct ddjvu_document_s
    struct ddjvu_page_s
    struct ddjvu_format_s
    struct ddjvu_rect_s
    struct ddjvu_rectmapper_s

    ctypedef ddjvu_context_s ddjvu_context_t
    ctypedef ddjvu_message_s ddjvu_message_t
    ctypedef ddjvu_job_s ddjvu_job_t
    ctypedef ddjvu_document_s ddjvu_document_t
    ctypedef ddjvu_page_s ddjvu_page_t
    ctypedef ddjvu_format_s ddjvu_format_t
    ctypedef ddjvu_rect_s ddjvu_rect_t
    ctypedef ddjvu_rectmapper_s ddjvu_rectmapper_t

    ctypedef void (*ddjvu_message_callback_t)(ddjvu_context_t* context, void* closure) nogil

    ctypedef enum ddjvu_status_t:
        DDJVU_JOB_NOTSTARTED
        DDJVU_JOB_STARTED
        DDJVU_JOB_OK
        DDJVU_JOB_FAILED
        DDJVU_JOB_STOPPED

    ctypedef enum ddjvu_message_tag_t:
        DDJVU_ERROR
        DDJVU_INFO
        DDJVU_NEWSTREAM
        DDJVU_DOCINFO
        DDJVU_PAGEINFO
        DDJVU_RELAYOUT
        DDJVU_REDISPLAY
        DDJVU_CHUNK
        DDJVU_THUMBNAIL
        DDJVU_PROGRESS

    cdef struct ddjvu_message_any_s:
        ddjvu_message_tag_t tag
        ddjvu_context_t* context
        ddjvu_document_t* document
        ddjvu_page_t* page
        ddjvu_job_t* job
    ctypedef ddjvu_message_any_s ddjvu_message_any_t

    cdef struct ddjvu_message_error_s:
        ddjvu_message_any_t any
        char* message
        char* function
        char* filename
        int lineno

    cdef struct ddjvu_message_info_s:
        ddjvu_message_any_t any
        char* message

    cdef struct ddjvu_message_newstream_s:
        ddjvu_message_any_t any
        int streamid
        char* name
        char* url

    cdef struct ddjvu_message_docinfo_s:
        ddjvu_message_any_t any

    ctypedef enum ddjvu_document_type_t:
        DDJVU_DOCTYPE_UNKNOWN
        DDJVU_DOCTYPE_SINGLEPAGE
        DDJVU_DOCTYPE_BUNDLED
        DDJVU_DOCTYPE_INDIRECT
        DDJVU_DOCTYPE_OLD_BUNDLED
        DDJVU_DOCTYPE_OLD_INDEXED

    cdef struct ddjvu_fileinfo_s:
        char type
        int pageno
        int size
        char* id
        char* name
        char* title
    ctypedef ddjvu_fileinfo_s ddjvu_fileinfo_t

    cdef struct ddjvu_pageinfo_s:
        int width
        int height
        int dpi
        int rotation
        int version
    ctypedef ddjvu_pageinfo_s ddjvu_pageinfo_t

    cdef struct ddjvu_message_pageinfo_s:
        ddjvu_message_any_t any

    cdef struct ddjvu_message_relayout_s:
        ddjvu_message_any_t any

    cdef struct ddjvu_message_redisplay_s:
        ddjvu_message_any_t any

    cdef struct ddjvu_message_chunk_s:
        ddjvu_message_any_t any
    char* chunkid

    ctypedef enum ddjvu_page_type_t:
        DDJVU_PAGETYPE_UNKNOWN
        DDJVU_PAGETYPE_BITONAL
        DDJVU_PAGETYPE_PHOTO
        DDJVU_PAGETYPE_COMPOUND

    ctypedef enum ddjvu_page_rotation_t:
        DDJVU_ROTATE_0
        DDJVU_ROTATE_90
        DDJVU_ROTATE_180
        DDJVU_ROTATE_270

    ctypedef enum ddjvu_render_mode_t:
        DDJVU_RENDER_COLOR
        DDJVU_RENDER_BLACK
        DDJVU_RENDER_COLORONLY
        DDJVU_RENDER_MASKONLY
        DDJVU_RENDER_BACKGROUND
        DDJVU_RENDER_FOREGROUND

    cdef struct ddjvu_rect_s:
        int x, y
        unsigned int w, h

    ctypedef enum ddjvu_format_style_t:
        DDJVU_FORMAT_BGR24
        DDJVU_FORMAT_RGB24
        DDJVU_FORMAT_RGBMASK16
        DDJVU_FORMAT_RGBMASK32
        DDJVU_FORMAT_GREY8
        DDJVU_FORMAT_PALETTE8
        DDJVU_FORMAT_MSBTOLSB
        DDJVU_FORMAT_LSBTOMSB

    cdef struct ddjvu_message_thumbnail_s:
        ddjvu_message_any_t any
        int pagenum

    cdef struct ddjvu_message_progress_s:
        ddjvu_message_any_t any
        ddjvu_status_t status
        int percent

    cdef union ddjvu_message_s:
        ddjvu_message_any_s m_any
        ddjvu_message_error_s m_error
        ddjvu_message_info_s m_info
        ddjvu_message_newstream_s m_newstream
        ddjvu_message_docinfo_s m_docinfo
        ddjvu_message_pageinfo_s m_pageinfo
        ddjvu_message_chunk_s m_chunk
        ddjvu_message_relayout_s m_relayout
        ddjvu_message_redisplay_s m_redisplay
        ddjvu_message_thumbnail_s m_thumbnail
        ddjvu_message_progress_s m_progress

cdef class Context

cdef class Document

cdef class DocumentExtension:
    cdef Document _document

cdef class DocumentPages(DocumentExtension):
    pass

cdef class DocumentFiles(DocumentExtension):
    cdef object _page_map

cdef class Document:
    cdef ddjvu_document_t* ddjvu_document
    cdef Context _context
    cdef DocumentPages _pages
    cdef DocumentFiles _files
    cdef object _queue
    cdef object _condition
    cdef object __weakref__
    cdef object __init(self, Context context, ddjvu_document_t* ddjvu_document)
    cdef object __clear(self)

cdef class _SexprWrapper:
    cdef object _document_weakref
    cdef cexpr_t _cexpr

cdef class DocumentOutline(DocumentExtension):
    cdef _SexprWrapper _sexpr
    cdef object _update_sexpr(self)

cdef class Annotations:
    cdef _SexprWrapper _sexpr
    cdef object _update_sexpr(self)
    cdef Document _document

cdef class DocumentAnnotations(Annotations):
    cdef int _compat

cdef class Hyperlinks:
    cdef object _sexpr

cdef class Metadata:
    cdef Annotations _annotations
    cdef object _keys

cdef class File:
    cdef int _n
    cdef int _have_info
    cdef ddjvu_fileinfo_t ddjvu_fileinfo
    cdef Document _document
    cdef object _get_info(self)

cdef class Page:
    cdef Document _document
    cdef ddjvu_pageinfo_t ddjvu_pageinfo
    cdef int _have_info
    cdef int _n
    cdef object _get_info(self)

cdef class PageAnnotations(Annotations):
    cdef Page _page

cdef class PageText:
    cdef Page _page
    cdef object _details
    cdef _SexprWrapper _sexpr
    cdef object _update_sexpr(self)

cdef class Context:
    cdef ddjvu_context_t* ddjvu_context
    cdef object _queue

cdef class PixelFormat:
    cdef ddjvu_format_t* ddjvu_format
    cdef int _bpp
    cdef int _dither_bpp
    cdef int _row_order
    cdef int _y_direction
    cdef double _gamma

cdef class PixelFormatRgb(PixelFormat):
    cdef int _rgb

cdef class PixelFormatRgbMask(PixelFormat):
    cdef unsigned int _params[4]

cdef class PixelFormatGrey(PixelFormat):
    pass

cdef class PixelFormatPalette(PixelFormat):
    cdef unsigned int _palette[216]

cdef class PixelFormatPackedBits(PixelFormat):
    cdef int _little_endian
    pass

cdef class Job:
    cdef Context _context
    cdef ddjvu_job_t* ddjvu_job
    cdef object _queue
    cdef object _condition
    cdef object __init(self, Context context, ddjvu_job_t *ddjvu_job)
    cdef object __clear(self)
    cdef object __weakref__

cdef class PageJob(Job):
    pass

cdef class SaveJob(Job):
    cdef object _file

cdef class DocumentDecodingJob(Job):
    cdef object _document
    cdef object __init_ddj(self, Document document)

cdef class AffineTransform:
    cdef ddjvu_rectmapper_t* ddjvu_rectmapper

cdef class Message:
    cdef ddjvu_message_t* ddjvu_message
    cdef Context _context
    cdef Document _document
    cdef PageJob _page_job
    cdef Job _job
    cdef object __init(self)

cdef class ErrorMessage(Message):
    cdef object _message
    cdef object _location

cdef class InfoMessage(Message):
    cdef object _message

cdef class Stream:
    cdef int _streamid
    cdef int _open
    cdef Document _document

cdef class NewStreamMessage(Message):
    cdef object _name
    cdef object _uri
    cdef Stream _stream

cdef class DocInfoMessage(Message):
    pass

cdef class PageInfoMessage(Message):
    pass

cdef class ChunkMessage(Message):
    pass

cdef class RelayoutMessage(ChunkMessage):
    pass

cdef class RedisplayMessage(ChunkMessage):
    pass

cdef class ThumbnailMessage(Message):
    cdef int _page_no

cdef class ProgressMessage(Message):
    cdef int _percent
    cdef ddjvu_status_t _status

cdef class Thumbnail:
    cdef Page _page

# vim:ts=4 sts=4 sw=4 et ft=pyrex



# encoding=UTF-8

# Copyright © 2007-2016 Jakub Wilk <jwilk@jwilk.net>
#
# This file is part of python-djvulibre.
#
# python-djvulibre is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# python-djvulibre is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.

import array
import errno
import os
import shutil
import subprocess
import sys
import tempfile

from djvu.decode import (
    AffineTransform,
    Context,
    DDJVU_VERSION,
    DOCUMENT_TYPE_BUNDLED,
    DOCUMENT_TYPE_SINGLE_PAGE,
    DocInfoMessage,
    Document,
    DocumentAnnotations,
    DocumentDecodingJob,
    DocumentOutline,
    ErrorMessage,
    File,
    FileUri,
    Hyperlinks,
    Job,
    JobFailed,
    JobOK,
    Message,
    Metadata,
    NewStreamMessage,
    NotAvailable,
    PAGE_TYPE_BITONAL,
    Page,
    PageAnnotations,
    PageJob,
    PageText,
    PixelFormat,
    PixelFormatGrey,
    PixelFormatPackedBits,
    PixelFormatPalette,
    PixelFormatRgb,
    PixelFormatRgbMask,
    RENDER_COLOR,
    SaveJob,
    Stream,
    TEXT_DETAILS_ALL,
    TEXT_DETAILS_CHARACTER,
    TEXT_DETAILS_COLUMN,
    TEXT_DETAILS_LINE,
    TEXT_DETAILS_PAGE,
    TEXT_DETAILS_PARAGRAPH,
    TEXT_DETAILS_REGION,
    TEXT_DETAILS_WORD,
    ThumbnailMessage,
    __version__,
)
from djvu.sexpr import (
    Expression,
    Symbol,
)

from tools import (
    assert_equal,
    assert_false,
    assert_is,
    assert_is_instance,
    assert_list_equal,
    assert_multi_line_equal,
    assert_raises,
    assert_raises_regex,
    assert_raises_str,
    assert_repr,
    assert_true,
    SkipTest,
    skip_unless_c_messages,
    skip_unless_command_exists,
    skip_unless_translation_exists,
    interim_locale,
    locale_encoding,
    wildcard_import,
    # Python 2/3 compat:
    b,
    maxsize,
    py3k,
    u,
    unicode,
)

images = os.path.join(os.path.dirname(__file__), 'images', '')

if sys.version_info >= (3, 2):
    array_tobytes = array.array.tobytes
else:
    array_tobytes = array.array.tostring

def run(*cmd, **kwargs):
    stdin = kwargs.pop('stdin', None)
    env = dict(os.environ)
    for key, value in kwargs.items():
        if key.isupper():
            env[key] = value
            continue
        raise TypeError('{key!r} is an invalid keyword argument for this function'.format(key=key))
    kwargs = dict(
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=env,
    )
    if stdin is not None:
        kwargs.update(stdin=subprocess.PIPE)
    child = subprocess.Popen(list(cmd), **kwargs)
    (stdout, stderr) = child.communicate(stdin)
    if child.returncode != 0:
        raise subprocess.CalledProcessError(child.returncode, cmd[0])
    return (stdout, stderr)

def create_djvu(commands='', sexpr=''):
    skip_unless_command_exists('djvused')
    if sexpr:
        commands += '\nset-ant\n{sexpr}\n.\n'.format(sexpr=sexpr)
    file = tempfile.NamedTemporaryFile(prefix='test', suffix='djvu')
    file.seek(0)
    file.write(
        b'\x41\x54\x26\x54\x46\x4f\x52\x4d\x00\x00\x00\x22\x44\x4a\x56\x55'
        b'\x49\x4e\x46\x4f\x00\x00\x00\x0a\x00\x01\x00\x01\x18\x00\x2c\x01'
        b'\x16\x01\x53\x6a\x62\x7a\x00\x00\x00\x04\xbc\x73\x1b\xd7'
    )
    file.flush()
    (stdout, stderr) = run('djvused', '-s', file.name, stdin=commands.encode(locale_encoding))
    assert_equal(stdout, ''.encode(locale_encoding))
    assert_equal(stderr, ''.encode(locale_encoding))
    return file

def test_context_cache():
    context = Context()
    assert_equal(context.cache_size, 10 << 20)
    for n in -100, 0, 1 << 31:
        with assert_raises_str(ValueError, '0 < cache_size < (2 ** 31) must be satisfied'):
            context.cache_size = n
    with assert_raises_str(ValueError, '0 < cache_size < (2 ** 31) must be satisfied'):
        context.cache_size = 0
    n = 1
    while n < (1 << 31):
        context.cache_size = n
        assert_equal(context.cache_size, n)
        n = (n + 1) * 2 - 1
    context.clear_cache()

class test_documents:

    def test_bad_new(self):
        with assert_raises_str(TypeError, "cannot create 'djvu.decode.Document' instances"):
            Document()

    def test_nonexistent(self):
        path = '__nonexistent__'
        try:
            os.stat(path)
        except OSError as ex:
            c_message = ex.args[1]
        else:
            raise OSError(errno.EEXIST, os.strerror(errno.EEXIST), path)
        c_message.encode('ASCII')
        skip_unless_c_messages()
        context = Context()
        with assert_raises(JobFailed):
            context.new_document(FileUri(path))
        message = context.get_message()
        assert_equal(type(message), ErrorMessage)
        assert_equal(type(message.message), unicode)
        assert_equal(
            message.message,
            "[1-11711] Failed to open '{path}': {msg}.".format(path=path, msg=c_message)
        )
        assert_equal(str(message), message.message)
        assert_equal(unicode(message), message.message)

    def test_nonexistent_ja(self):
        skip_unless_c_messages()
        skip_unless_translation_exists('ja_JP.UTF-8')
        path = '__nonexistent__'
        context = Context()
        try:
            with interim_locale(LC_ALL='ja_JP.UTF-8'):
                os.stat(path)
        except OSError as ex:
            c_message = ex.args[1]
        else:
            raise OSError(errno.EEXIST, os.strerror(errno.EEXIST), path)
        try:
            c_message.encode('ASCII')
        except UnicodeError:
            pass
        else:
            raise AssertionError(
                'ja_JP error message is ASCII-only: {msg!r}'.format(msg=c_message)
            )
        with interim_locale(LC_ALL='ja_JP.UTF-8'):
            with assert_raises(JobFailed):
                context.new_document(FileUri(path))
            message = context.get_message()
            assert_equal(type(message), ErrorMessage)
            assert_equal(type(message.message), unicode)
            assert_equal(
                message.message,
                u("[1-11711] Failed to open '{path}': {msg}.".format(path=path, msg=c_message))
            )
            assert_equal(
                str(message),
                "[1-11711] Failed to open '{path}': {msg}.".format(path=path, msg=c_message)
            )
            assert_equal(unicode(message), message.message)

    def test_new_document(self):
        context = Context()
        document = context.new_document(FileUri(images + 'test1.djvu'))
        assert_equal(type(document), Document)
        message = document.get_message()
        assert_equal(type(message), DocInfoMessage)
        assert_true(document.decoding_done)
        assert_false(document.decoding_error)
        assert_equal(document.decoding_status, JobOK)
        assert_equal(document.type, DOCUMENT_TYPE_SINGLE_PAGE)
        assert_equal(len(document.pages), 1)
        assert_equal(len(document.files), 1)
        decoding_job = document.decoding_job
        assert_true(decoding_job.is_done)
        assert_false(decoding_job.is_error)
        assert_equal(decoding_job.status, JobOK)
        file = document.files[0]
        assert_is(type(file), File)
        assert_is(file.document, document)
        assert_is(file.get_info(), None)
        assert_equal(file.type, 'P')
        assert_equal(file.n_page, 0)
        page = file.page
        assert_equal(type(page), Page)
        assert_is(page.document, document)
        assert_equal(page.n, 0)
        assert_is(file.size, None)
        assert_equal(file.id, u('test1.djvu'))
        assert_equal(type(file.id), unicode)
        assert_equal(file.name, u('test1.djvu'))
        assert_equal(type(file.name), unicode)
        assert_equal(file.title, u('test1.djvu'))
        assert_equal(type(file.title), unicode)
        dump = document.files[0].dump
        assert_equal(type(dump), unicode)
        assert_equal(
            [line for line in dump.splitlines()], [
                u('  FORM:DJVU [83] '),
                u('    INFO [10]         DjVu 64x48, v24, 300 dpi, gamma=2.2'),
                u('    Sjbz [53]         JB2 bilevel data'),
            ]
        )
        page = document.pages[0]
        assert_equal(type(page), Page)
        assert_is(page.document, document)
        assert_is(page.get_info(), None)
        assert_equal(page.width, 64)
        assert_equal(page.height, 48)
        assert_equal(page.size, (64, 48))
        assert_equal(page.dpi, 300)
        assert_equal(page.rotation, 0)
        assert_equal(page.version, 24)
        file = page.file
        assert_equal(type(file), File)
        assert_equal(file.id, u('test1.djvu'))
        assert_equal(type(file.id), unicode)
        dump = document.files[0].dump
        assert_equal(type(dump), unicode)
        assert_equal(
            [line for line in dump.splitlines()], [
                u('  FORM:DJVU [83] '),
                u('    INFO [10]         DjVu 64x48, v24, 300 dpi, gamma=2.2'),
                u('    Sjbz [53]         JB2 bilevel data'),
            ]
        )
        assert_is(document.get_message(wait=False), None)
        assert_is(context.get_message(wait=False), None)
        with assert_raises_str(IndexError, 'file number out of range'):
            document.files[-1].get_info()
        assert_is(document.get_message(wait=False), None)
        assert_is(context.get_message(wait=False), None)
        with assert_raises_str(IndexError, 'page number out of range'):
            document.pages[-1]
        with assert_raises_str(IndexError, 'page number out of range'):
            document.pages[1]
        assert_is(document.get_message(wait=False), None)
        assert_is(context.get_message(wait=False), None)

    def test_save(self):
        skip_unless_command_exists('djvudump')
        context = Context()
        original_filename = images + 'test0.djvu'
        document = context.new_document(FileUri(original_filename))
        message = document.get_message()
        assert_equal(type(message), DocInfoMessage)
        assert_true(document.decoding_done)
        assert_false(document.decoding_error)
        assert_equal(document.decoding_status, JobOK)
        assert_equal(document.type, DOCUMENT_TYPE_BUNDLED)
        assert_equal(len(document.pages), 2)
        assert_equal(len(document.files), 3)
        (stdout0, stderr0) = run('djvudump', original_filename, LC_ALL='C')
        assert_equal(stderr0, b(''))
        stdout0 = stdout0.replace(b('\r\n'), b('\n'))
        tmpdir = tempfile.mkdtemp()
        try:
            tmp = open(os.path.join(tmpdir, 'tmp.djvu'), 'wb')
            job = document.save(tmp)
            assert_equal(type(job), SaveJob)
            assert_true(job.is_done)
            assert_false(job.is_error)
            tmp.close()
            (stdout, stderr) = run('djvudump', tmp.name, LC_ALL='C')
            assert_equal(stderr, b(''))
            stdout = stdout.replace(b('\r\n'), b('\n'))
            assert_equal(stdout, stdout0)
        finally:
            shutil.rmtree(tmpdir)
            tmp = None
        tmpdir = tempfile.mkdtemp()
        try:
            tmp = open(os.path.join(tmpdir, 'tmp.djvu'), 'wb')
            job = document.save(tmp, pages=(0,))
            assert_equal(type(job), SaveJob)
            assert_true(job.is_done)
            assert_false(job.is_error)
            tmp.close()
            stdout, stderr = run('djvudump', tmp.name, LC_ALL='C')
            assert_equal(stderr, b(''))
            stdout = stdout.replace(b('\r\n'), b('\n'))
            stdout0 = stdout0.split(b('\n'))
            stdout = stdout.split(b('\n'))
            stdout[4] = stdout[4].replace(b(' (1)'), b(''))
            assert_equal(len(stdout), 10)
            assert_equal(stdout[3:-1], stdout0[4:10])
            assert_equal(stdout[-1], b(''))
        finally:
            shutil.rmtree(tmpdir)
            tmp = None
        tmpdir = tempfile.mkdtemp()
        try:
            tmpfname = os.path.join(tmpdir, 'index.djvu')
            job = document.save(indirect=tmpfname)
            assert_equal(type(job), SaveJob)
            assert_true(job.is_done)
            assert_false(job.is_error)
            (stdout, stderr) = run('djvudump', tmpfname, LC_ALL='C')
            assert_equal(stderr, b(''))
            stdout = stdout.replace(b('\r\n'), b('\n'))
            stdout = stdout.split(b('\n'))
            stdout0 = (
                [b('      shared_anno.iff -> shared_anno.iff')] +
                [b('      p{n:04}.djvu -> p{n:04}.djvu'.format(n=n)) for n in range(1, 3)]
            )
            assert_equal(len(stdout), 7)
            assert_equal(stdout[2:-2], stdout0)
            assert_equal(stdout[-1], b(''))
        finally:
            shutil.rmtree(tmpdir)
        tmpdir = tempfile.mkdtemp()
        try:
            tmpfname = os.path.join(tmpdir, 'index.djvu')
            job = document.save(indirect=tmpfname, pages=(0,))
            assert_equal(type(job), SaveJob)
            assert_true(job.is_done)
            assert_false(job.is_error)
            (stdout, stderr) = run('djvudump', tmpfname, LC_ALL='C')
            stdout = stdout.replace(b('\r\n'), b('\n'))
            assert_equal(stderr, b(''))
            stdout = stdout.split(b('\n'))
            assert_equal(len(stdout), 5)
            assert_equal(stdout[2], b('      shared_anno.iff -> shared_anno.iff'))
            assert_equal(stdout[3], b('      p0001.djvu -> p0001.djvu'))
            assert_equal(stdout[-1], b(''))
        finally:
            shutil.rmtree(tmpdir)

    def test_export_ps(self):
        skip_unless_command_exists('ps2ascii')
        context = Context()
        document = context.new_document(FileUri(images + 'test0.djvu'))
        message = document.get_message()
        assert_equal(type(message), DocInfoMessage)
        assert_true(document.decoding_done)
        assert_false(document.decoding_error)
        assert_equal(document.decoding_status, JobOK)
        assert_equal(document.type, DOCUMENT_TYPE_BUNDLED)
        assert_equal(len(document.pages), 2)
        assert_equal(len(document.files), 3)
        tmp = tempfile.NamedTemporaryFile()
        try:
            job = document.export_ps(tmp.file)
            assert_equal(type(job), SaveJob)
            assert_true(job.is_done)
            assert_false(job.is_error)
            stdout, stderr = run('ps2ascii', tmp.name, LC_ALL='C')
            assert_equal(stderr, b(''))
            assert_equal(stdout, b('\x0c') * 2)
        finally:
            tmp.close()

        tmp = tempfile.NamedTemporaryFile()
        try:
            job = document.export_ps(tmp.file, pages=(0,), text=True)
            assert_equal(type(job), SaveJob)
            assert_true(job.is_done)
            assert_false(job.is_error)
            stdout, stderr = run('ps2ascii', tmp.name, LC_ALL='C')
            assert_equal(stderr, b(''))
            stdout = stdout.decode('ASCII')
            stdout = ' '.join(stdout.split())
            expected = '''
                1 Lorem ipsum
                Optio reprehenderit molestias amet aliquam, similique doloremque fuga labore
                voluptatum voluptatem, commodi culpa voluptas, officia tenetur expedita quidem
                hic repellat molestiae quis accusamus dolores repudiandae, quidem in ad
                voluptas eligendi maiores placeat ex consectetur at tenetur amet.
                1
            '''
            expected = ' '.join(expected.split())
            assert_multi_line_equal(stdout, expected)
        finally:
            del tmp

class test_pixel_formats():

    def test_bad_new(self):
        with assert_raises_str(TypeError, "cannot create 'djvu.decode.PixelFormat' instances"):
            PixelFormat()

    def test_rgb(self):
        pf = PixelFormatRgb()
        assert_repr(pf, "djvu.decode.PixelFormatRgb(byte_order = 'RGB', bpp = 24)")
        pf = PixelFormatRgb('RGB')
        assert_repr(pf, "djvu.decode.PixelFormatRgb(byte_order = 'RGB', bpp = 24)")
        pf = PixelFormatRgb('BGR')
        assert_repr(pf, "djvu.decode.PixelFormatRgb(byte_order = 'BGR', bpp = 24)")

    def test_rgb_mask(self):
        pf = PixelFormatRgbMask(0xff, 0xf00, 0x1f000, 0, 16)
        assert_repr(pf, "djvu.decode.PixelFormatRgbMask(red_mask = 0x00ff, green_mask = 0x0f00, blue_mask = 0xf000, xor_value = 0x0000, bpp = 16)")
        pf = PixelFormatRgbMask(0xff000000, 0xff0000, 0xff00, 0xff, 32)
        assert_repr(pf, "djvu.decode.PixelFormatRgbMask(red_mask = 0xff000000, green_mask = 0x00ff0000, blue_mask = 0x0000ff00, xor_value = 0x000000ff, bpp = 32)")

    def test_grey(self):
        pf = PixelFormatGrey()
        assert_repr(pf, "djvu.decode.PixelFormatGrey(bpp = 8)")

    def test_palette(self):
        with assert_raises(KeyError) as ecm:
            pf = PixelFormatPalette({})
        assert_equal(
            ecm.exception.args,
            ((0, 0, 0),)
        )
        data = dict(((i, j, k), i + 7 * j + 37 + k) for i in range(6) for j in range(6) for k in range(6))
        pf = PixelFormatPalette(data)
        data_repr = ', '.join(
            '{k!r}: 0x{v:02x}'.format(k=k, v=v) for k, v in sorted(data.items())
        )
        assert_equal(
            repr(pf),
            'djvu.decode.PixelFormatPalette({{{data}}}, bpp = 8)'.format(data=data_repr)
        )

    def test_packed_bits(self):
        pf = PixelFormatPackedBits('<')
        assert_repr(pf, "djvu.decode.PixelFormatPackedBits('<')")
        assert_equal(pf.bpp, 1)
        pf = PixelFormatPackedBits('>')
        assert_repr(pf, "djvu.decode.PixelFormatPackedBits('>')")
        assert_equal(pf.bpp, 1)

class test_page_jobs():

    def test_bad_new(self):
        with assert_raises_str(TypeError, "cannot create 'djvu.decode.PageJob' instances"):
            PageJob()

    def test_decode(self):
        context = Context()
        document = context.new_document(FileUri(images + 'test1.djvu'))
        message = document.get_message()
        assert_equal(type(message), DocInfoMessage)
        page_job = document.pages[0].decode()
        assert_true(page_job.is_done)
        assert_equal(type(page_job), PageJob)
        assert_true(page_job.is_done)
        assert_false(page_job.is_error)
        assert_equal(page_job.status, JobOK)
        assert_equal(page_job.width, 64)
        assert_equal(page_job.height, 48)
        assert_equal(page_job.size, (64, 48))
        assert_equal(page_job.dpi, 300)
        assert_equal(page_job.gamma, 2.2)
        assert_equal(page_job.version, 24)
        assert_equal(page_job.type, PAGE_TYPE_BITONAL)
        assert_equal((page_job.rotation, page_job.initial_rotation), (0, 0))
        with assert_raises_str(ValueError, 'rotation must be equal to 0, 90, 180, or 270'):
            page_job.rotation = 100
        page_job.rotation = 180
        assert_equal((page_job.rotation, page_job.initial_rotation), (180, 0))
        del page_job.rotation
        assert_equal((page_job.rotation, page_job.initial_rotation), (0, 0))

        with assert_raises_str(ValueError, 'page_rect width/height must be a positive integer'):
            page_job.render(RENDER_COLOR, (0, 0, -1, -1), (0, 0, 10, 10), PixelFormatRgb())

        with assert_raises_str(ValueError, 'render_rect width/height must be a positive integer'):
            page_job.render(RENDER_COLOR, (0, 0, 10, 10), (0, 0, -1, -1), PixelFormatRgb())

        with assert_raises_str(ValueError, 'render_rect must be inside page_rect'):
            page_job.render(RENDER_COLOR, (0, 0, 10, 10), (2, 2, 10, 10), PixelFormatRgb())

        with assert_raises_str(ValueError, 'row_alignment must be a positive integer'):
            page_job.render(RENDER_COLOR, (0, 0, 10, 10), (0, 0, 10, 10), PixelFormatRgb(), -1)

        with assert_raises_regex(MemoryError, r'\AUnable to allocate [0-9]+ bytes for an image memory\Z'):
            x = int((maxsize // 2) ** 0.5)
            page_job.render(RENDER_COLOR, (0, 0, x, x), (0, 0, x, x), PixelFormatRgb(), 8)

        s = page_job.render(RENDER_COLOR, (0, 0, 10, 10), (0, 0, 4, 4), PixelFormatGrey(), 1)
        assert_equal(s, b'\xff\xff\xff\xff\xff\xff\xff\xef\xff\xff\xff\xa4\xff\xff\xff\xb8')

        buffer = array.array('B', b'\0')
        with assert_raises_str(ValueError, 'Image buffer is too small (16 > 1)'):
            page_job.render(RENDER_COLOR, (0, 0, 10, 10), (0, 0, 4, 4), PixelFormatGrey(), 1, buffer)

        buffer = array.array('B', b'\0' * 16)
        assert_is(page_job.render(RENDER_COLOR, (0, 0, 10, 10), (0, 0, 4, 4), PixelFormatGrey(), 1, buffer), buffer)
        s = array_tobytes(buffer)
        assert_equal(s, b'\xff\xff\xff\xff\xff\xff\xff\xef\xff\xff\xff\xa4\xff\xff\xff\xb8')

class test_thumbnails:

    def test(self):
        context = Context()
        document = context.new_document(FileUri(images + 'test1.djvu'))
        message = document.get_message()
        assert_equal(type(message), DocInfoMessage)
        thumbnail = document.pages[0].thumbnail
        assert_equal(thumbnail.status, JobOK)
        assert_equal(thumbnail.calculate(), JobOK)
        message = document.get_message()
        assert_equal(type(message), ThumbnailMessage)
        assert_equal(message.thumbnail.page.n, 0)
        (w, h, r), pixels = thumbnail.render((5, 5), PixelFormatGrey(), dry_run=True)
        assert_equal((w, h, r), (5, 3, 5))
        assert_is(pixels, None)
        (w, h, r), pixels = thumbnail.render((5, 5), PixelFormatGrey())
        assert_equal((w, h, r), (5, 3, 5))
        assert_equal(pixels[:15], b'\xff\xeb\xa7\xf2\xff\xff\xbf\x86\xbe\xff\xff\xe7\xd6\xe7\xff')
        buffer = array.array('B', b'\0')
        with assert_raises_str(ValueError, 'Image buffer is too small (25 > 1)'):
            (w, h, r), pixels = thumbnail.render((5, 5), PixelFormatGrey(), buffer=buffer)
        buffer = array.array('B', b'\0' * 25)
        (w, h, r), pixels = thumbnail.render((5, 5), PixelFormatGrey(), buffer=buffer)
        assert_is(pixels, buffer)
        s = array_tobytes(buffer[:15])
        assert_equal(s, b'\xff\xeb\xa7\xf2\xff\xff\xbf\x86\xbe\xff\xff\xe7\xd6\xe7\xff')

def test_jobs():

    with assert_raises_str(TypeError, "cannot create 'djvu.decode.Job' instances"):
        Job()

    with assert_raises_str(TypeError, "cannot create 'djvu.decode.DocumentDecodingJob' instances"):
        DocumentDecodingJob()

class test_affine_transforms():

    def test_bad_args(self):
        with assert_raises_str(ValueError, 'need more than 2 values to unpack'):
            AffineTransform((1, 2), (3, 4, 5))

    def test1(self):
        af = AffineTransform((0, 0, 10, 10), (17, 42, 42, 100))
        assert_equal(type(af), AffineTransform)
        assert_equal(af((0, 0)), (17, 42))
        assert_equal(af((0, 10)), (17, 142))
        assert_equal(af((10, 0)), (59, 42))
        assert_equal(af((10, 10)), (59, 142))
        assert_equal(af((0, 0, 10, 10)), (17, 42, 42, 100))
        assert_equal(af(x for x in (0, 0, 10, 10)), (17, 42, 42, 100))
        assert_equal(af.apply((123, 456)), af((123, 456)))
        assert_equal(af.apply((12, 34, 56, 78)), af((12, 34, 56, 78)))
        assert_equal(af.inverse((17, 42)), (0, 0))
        assert_equal(af.inverse((17, 142)), (0, 10))
        assert_equal(af.inverse((59, 42)), (10, 0))
        assert_equal(af.inverse((59, 142)), (10, 10))
        assert_equal(af.inverse((17, 42, 42, 100)), (0, 0, 10, 10))
        assert_equal(af.inverse(x for x in (17, 42, 42, 100)), (0, 0, 10, 10))
        assert_equal(af.inverse(af((234, 567))), (234, 567))
        assert_equal(af.inverse(af((23, 45, 67, 78))), (23, 45, 67, 78))

class test_messages():

    def test_bad_new(self):
        with assert_raises_str(TypeError, "cannot create 'djvu.decode.Message' instances"):
            Message()

class test_streams:

    def test_bad_new(self):
        with assert_raises_str(TypeError, "Argument 'document' has incorrect type (expected djvu.decode.Document, got NoneType)"):
            Stream(None, 42)

    def test(self):
        context = Context()
        document = context.new_document('dummy://dummy.djvu')
        message = document.get_message()
        assert_equal(type(message), NewStreamMessage)
        assert_equal(message.name, 'dummy.djvu')
        assert_equal(message.uri, 'dummy://dummy.djvu')
        assert_equal(type(message.stream), Stream)
        with assert_raises(NotAvailable):
            document.outline.sexpr
        with assert_raises(NotAvailable):
            document.annotations.sexpr
        with assert_raises(NotAvailable):
            document.pages[0].text.sexpr
        with assert_raises(NotAvailable):
            document.pages[0].annotations.sexpr
        try:
            with open(images + 'test1.djvu', 'rb') as fp:
                message.stream.write(fp.read())
        finally:
            message.stream.close()
        with assert_raises_str(IOError, 'I/O operation on closed file'):
            message.stream.write(b('eggs'))
        message = document.get_message()
        assert_equal(type(message), DocInfoMessage)
        outline = document.outline
        outline.wait()
        x = outline.sexpr
        assert_equal(x, Expression([]))
        anno = document.annotations
        anno.wait()
        x = anno.sexpr
        assert_equal(x, Expression([]))
        text = document.pages[0].text
        text.wait()
        x = text.sexpr
        assert_equal(x, Expression([]))
        anno = document.pages[0].annotations
        anno.wait()
        x = anno.sexpr
        assert_equal(x, Expression([]))

def test_metadata():

    model_metadata = {
        'English': 'eggs',
        u('Русский'): u('яйца'),
    }
    meta = '\n'.join(u('|{k}| {v}').format(k=k, v=v) for k, v in model_metadata.items())
    test_script = u('set-meta\n{meta}\n.\n').format(meta=meta)
    try:
        test_file = create_djvu(test_script)
    except UnicodeEncodeError:
        raise SkipTest('you need to run this test with LC_CTYPE=C or LC_CTYPE=<lang>.UTF-8')
    try:
        context = Context()
        document = context.new_document(FileUri(test_file.name))
        message = document.get_message()
        assert_equal(type(message), DocInfoMessage)
        annotations = document.annotations
        assert_equal(type(annotations), DocumentAnnotations)
        annotations.wait()
        metadata = annotations.metadata
        assert_equal(type(metadata), Metadata)
        assert_equal(len(metadata), len(model_metadata))
        assert_equal(sorted(metadata), sorted(model_metadata))
        if not py3k:
            assert_equal(sorted(metadata.iterkeys()), sorted(model_metadata.iterkeys()))
        assert_equal(sorted(metadata.keys()), sorted(model_metadata.keys()))
        if not py3k:
            assert_equal(sorted(metadata.itervalues()), sorted(model_metadata.itervalues()))
        assert_equal(sorted(metadata.values()), sorted(model_metadata.values()))
        if not py3k:
            assert_equal(sorted(metadata.iteritems()), sorted(model_metadata.iteritems()))
        assert_equal(sorted(metadata.items()), sorted(model_metadata.items()))
        for k in metadata:
            assert_equal(type(k), unicode)
            assert_equal(type(metadata[k]), unicode)
        for k in None, 42, '+'.join(model_metadata):
            with assert_raises(KeyError) as ecm:
                metadata[k]
            assert_equal(ecm.exception.args, (k,))
    finally:
        test_file.close()

class test_sexpr:

    def test(self):
        context = Context()
        document = context.new_document(FileUri(images + 'test0.djvu'))
        assert_equal(type(document), Document)
        message = document.get_message()
        assert_equal(type(message), DocInfoMessage)
        anno = DocumentAnnotations(document, shared=False)
        assert_equal(type(anno), DocumentAnnotations)
        anno.wait()
        x = anno.sexpr
        assert_equal(x, Expression([]))
        anno = document.annotations
        assert_equal(type(anno), DocumentAnnotations)
        anno.wait()
        assert_is(anno.background_color, None)
        assert_is(anno.horizontal_align, None)
        assert_is(anno.vertical_align, None)
        assert_is(anno.mode, None)
        assert_is(anno.zoom, None)
        expected_metadata = [
            Symbol('metadata'),
            [Symbol('ModDate'), '2015-08-17 19:54:57+02:00'],
            [Symbol('CreationDate'), '2015-08-17 19:54:57+02:00'],
            [Symbol('Producer'), 'pdfTeX-1.40.16'],
            [Symbol('Creator'), 'LaTeX with hyperref package'],
            [Symbol('Author'), 'Jakub Wilk']
        ]
        expected_xmp = [
            Symbol('xmp'),
            '<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">'
            '<rdf:Description rdf:about="">'
                '<xmpMM:History xmlns:xmpMM="http://ns.adobe.com/xap/1.0/mm/"><rdf:Seq><rdf:li xmlns:stEvt="http://ns.adobe.com/xap/1.0/sType/ResourceEvent#" stEvt:action="converted" stEvt:parameters="from application/pdf to image/vnd.djvu" stEvt:softwareAgent="pdf2djvu 0.8.1 (DjVuLibre 3.5.27, Poppler 0.26.5, GraphicsMagick++ 1.3.21, GNOME XSLT 1.1.28, GNOME XML 2.9.2, PStreams 0.8.0)" stEvt:when="2015-08-17T17:54:58+00:00"/></rdf:Seq></xmpMM:History>'
                '<dc:creator xmlns:dc="http://purl.org/dc/elements/1.1/">Jakub Wilk</dc:creator>'
                '<dc:format xmlns:dc="http://purl.org/dc/elements/1.1/">image/vnd.djvu</dc:format>'
                '<pdf:Producer xmlns:pdf="http://ns.adobe.com/pdf/1.3/">pdfTeX-1.40.16</pdf:Producer>'
                '<xmp:CreatorTool xmlns:xmp="http://ns.adobe.com/xap/1.0/">LaTeX with hyperref package</xmp:CreatorTool>'
                '<xmp:CreateDate xmlns:xmp="http://ns.adobe.com/xap/1.0/">2015-08-17T19:54:57+02:00</xmp:CreateDate>'
                '<xmp:ModifyDate xmlns:xmp="http://ns.adobe.com/xap/1.0/">2015-08-17T19:54:57+02:00</xmp:ModifyDate>'
                '<xmp:MetadataDate xmlns:xmp="http://ns.adobe.com/xap/1.0/">2015-08-17T17:54:58+00:00</xmp:MetadataDate>'
            '</rdf:Description>'
            '</rdf:RDF>\n'
        ]
        assert_equal(
            anno.sexpr,
            Expression([expected_metadata, expected_xmp])
        )
        metadata = anno.metadata
        assert_equal(type(metadata), Metadata)
        hyperlinks = anno.hyperlinks
        assert_equal(type(hyperlinks), Hyperlinks)
        assert_equal(len(hyperlinks), 0)
        assert_equal(list(hyperlinks), [])
        outline = document.outline
        assert_equal(type(outline), DocumentOutline)
        outline.wait()
        assert_equal(outline.sexpr, Expression(
            [Symbol('bookmarks'),
                ['Lorem ipsum', '#p0001.djvu'],
                ['Hyperlinks', '#p0002.djvu',
                    ['local', '#p0002.djvu'],
                    ['remote', '#p0002.djvu']
                ]
            ]
        ))
        page = document.pages[1]
        anno = page.annotations
        assert_equal(type(anno), PageAnnotations)
        anno.wait()
        assert_is(anno.background_color, None)
        assert_is(anno.horizontal_align, None)
        assert_is(anno.vertical_align, None)
        assert_is(anno.mode, None)
        assert_is(anno.zoom, None)
        expected_hyperlinks = [
            [Symbol('maparea'), '#p0001.djvu', '', [Symbol('rect'), 520, 2502, 33, 42], [Symbol('border'), Symbol('#ff0000')]],
            [Symbol('maparea'), 'http://jwilk.net/', '', [Symbol('rect'), 458, 2253, 516, 49], [Symbol('border'), Symbol('#00ffff')]]
        ]
        assert_equal(
            anno.sexpr,
            Expression([expected_metadata, expected_xmp] + expected_hyperlinks)
        )
        page_metadata = anno.metadata
        assert_equal(type(page_metadata), Metadata)
        assert_equal(page_metadata.keys(), metadata.keys())
        assert_equal([page_metadata[k] == metadata[k] for k in metadata], [True, True, True, True, True])
        hyperlinks = anno.hyperlinks
        assert_equal(type(hyperlinks), Hyperlinks)
        assert_equal(len(hyperlinks), 2)
        assert_equal(
            list(hyperlinks),
            [Expression(h) for h in expected_hyperlinks]
        )
        text = page.text
        assert_equal(type(text), PageText)
        text.wait()
        text_s = text.sexpr
        text_s_detail = [PageText(page, details).sexpr for details in (TEXT_DETAILS_PAGE, TEXT_DETAILS_COLUMN, TEXT_DETAILS_REGION, TEXT_DETAILS_PARAGRAPH, TEXT_DETAILS_LINE, TEXT_DETAILS_WORD, TEXT_DETAILS_CHARACTER, TEXT_DETAILS_ALL)]
        assert_equal(text_s_detail[0], text_s_detail[1])
        assert_equal(text_s_detail[1], text_s_detail[2])
        assert_equal(text_s_detail[2], text_s_detail[3])
        assert_equal(
            text_s_detail[0],
            Expression(
                [Symbol('page'), 0, 0, 2550, 3300,
                    '2 Hyperlinks \n'
                    '2.1 local \n'
                    + u('→1 \n') +
                    '2.2 remote \nhttp://jwilk.net/ \n'
                    '2 \n'
                ]
            )
        )
        assert_equal(
            text_s_detail[4],
            Expression(
                [Symbol('page'), 0, 0, 2550, 3300,
                    [Symbol('line'), 462, 2712, 910, 2777, '2 Hyperlinks '],
                    [Symbol('line'), 462, 2599, 714, 2641, '2.1 local '],
                    [Symbol('line'), 464, 2505, 544, 2540, u('→1 ')],
                    [Symbol('line'), 462, 2358, 772, 2400, '2.2 remote '],
                    [Symbol('line'), 463, 2256, 964, 2298, 'http://jwilk.net/ '],
                    [Symbol('line'), 1260, 375, 1282, 409, '2 ']
                ]
            )
        )
        assert_equal(text_s_detail[5], text_s)
        assert_equal(text_s_detail[6], text_s)
        assert_equal(text_s_detail[7], text_s)
        assert_equal(
            text_s,
            Expression(
                [Symbol('page'), 0, 0, 2550, 3300,
                    [Symbol('line'), 462, 2712, 910, 2777, [Symbol('word'), 462, 2727, 495, 2776, '2'], [Symbol('word'), 571, 2712, 910, 2777, 'Hyperlinks']],
                    [Symbol('line'), 462, 2599, 714, 2641, [Symbol('word'), 462, 2599, 532, 2641, '2.1'], [Symbol('word'), 597, 2599, 714, 2640, 'local']],
                    [Symbol('line'), 464, 2505, 544, 2540, [Symbol('word'), 464, 2505, 544, 2540, u('→1')]],
                    [Symbol('line'), 462, 2358, 772, 2400, [Symbol('word'), 462, 2358, 535, 2400, '2.2'], [Symbol('word'), 598, 2358, 772, 2397, 'remote']],
                    [Symbol('line'), 463, 2256, 964, 2298, [Symbol('word'), 463, 2256, 964, 2298, 'http://jwilk.net/']],
                    [Symbol('line'), 1260, 375, 1282, 409, [Symbol('word'), 1260, 375, 1282, 409, '2']]
                ]
            )
        )
        with assert_raises_str(TypeError, 'details must be a symbol or none'):
            PageText(page, 'eggs')
        with assert_raises_str(ValueError, 'details must be equal to TEXT_DETAILS_PAGE, or TEXT_DETAILS_COLUMN, or TEXT_DETAILS_REGION, or TEXT_DETAILS_PARAGRAPH, or TEXT_DETAILS_LINE, or TEXT_DETAILS_WORD, or TEXT_DETAILS_CHARACTER or TEXT_DETAILS_ALL'):
            PageText(page, Symbol('eggs'))

def test_version():
    assert_is_instance(__version__, str)
    assert_is_instance(DDJVU_VERSION, int)

def test_wildcard_import():
    ns = wildcard_import('djvu.decode')
    assert_list_equal(
        sorted(ns.keys()), [
            'AffineTransform',
            'Annotations',
            'ChunkMessage',
            'Context',
            'DDJVU_VERSION',
            'DOCUMENT_TYPE_BUNDLED',
            'DOCUMENT_TYPE_INDIRECT',
            'DOCUMENT_TYPE_OLD_BUNDLED',
            'DOCUMENT_TYPE_OLD_INDEXED',
            'DOCUMENT_TYPE_SINGLE_PAGE',
            'DOCUMENT_TYPE_UNKNOWN',
            'DocInfoMessage',
            'Document',
            'DocumentAnnotations',
            'DocumentDecodingJob',
            'DocumentExtension',
            'DocumentFiles',
            'DocumentOutline',
            'DocumentPages',
            'ErrorMessage',
            'FILE_TYPE_INCLUDE',
            'FILE_TYPE_PAGE',
            'FILE_TYPE_THUMBNAILS',
            'File',
            'FileURI',
            'FileUri',
            'Hyperlinks',
            'InfoMessage',
            'Job',
            'JobDone',
            'JobException',
            'JobFailed',
            'JobNotDone',
            'JobNotStarted',
            'JobOK',
            'JobStarted',
            'JobStopped',
            'Message',
            'Metadata',
            'NewStreamMessage',
            'NotAvailable',
            'PAGE_TYPE_BITONAL',
            'PAGE_TYPE_COMPOUND',
            'PAGE_TYPE_PHOTO',
            'PAGE_TYPE_UNKNOWN',
            'PRINT_BOOKLET_NO',
            'PRINT_BOOKLET_RECTO',
            'PRINT_BOOKLET_VERSO',
            'PRINT_BOOKLET_YES',
            'PRINT_ORIENTATION_AUTO',
            'PRINT_ORIENTATION_LANDSCAPE',
            'PRINT_ORIENTATION_PORTRAIT',
            'Page',
            'PageAnnotations',
            'PageInfoMessage',
            'PageJob',
            'PageText',
            'PixelFormat',
            'PixelFormatGrey',
            'PixelFormatPackedBits',
            'PixelFormatPalette',
            'PixelFormatRgb',
            'PixelFormatRgbMask',
            'ProgressMessage',
            'RENDER_BACKGROUND',
            'RENDER_BLACK',
            'RENDER_COLOR',
            'RENDER_COLOR_ONLY',
            'RENDER_FOREGROUND',
            'RENDER_MASK_ONLY',
            'RedisplayMessage',
            'RelayoutMessage',
            'SaveJob',
            'Stream',
            'TEXT_DETAILS_ALL',
            'TEXT_DETAILS_CHARACTER',
            'TEXT_DETAILS_COLUMN',
            'TEXT_DETAILS_LINE',
            'TEXT_DETAILS_PAGE',
            'TEXT_DETAILS_PARAGRAPH',
            'TEXT_DETAILS_REGION',
            'TEXT_DETAILS_WORD',
            'Thumbnail',
            'ThumbnailMessage',
            'cmp_text_zone'
        ]
    )

# vim:ts=4 sts=4 sw=4 et



# encoding=UTF-8

# Copyright © 2008-2015 Jakub Wilk <jwilk@jwilk.net>
#
# This file is part of python-djvulibre.
#
# python-djvulibre is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# python-djvulibre is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.

from djvu.const import (
    TEXT_ZONE_CHARACTER,
    TEXT_ZONE_COLUMN,
    TEXT_ZONE_LINE,
    TEXT_ZONE_PAGE,
    TEXT_ZONE_PARAGRAPH,
    TEXT_ZONE_REGION,
    TEXT_ZONE_WORD,
    TextZoneType,
    get_text_zone_type,
)

from djvu.sexpr import (
    Symbol,
)

from tools import (
    assert_equal,
    assert_is,
    assert_is_instance,
    assert_list_equal,
    assert_not_equal,
    assert_raises_str,
    assert_repr,
    wildcard_import,
    # Python 2/3 compat:
    cmp,
)

class test_text_zones():

    zones = [
        TEXT_ZONE_PAGE,
        TEXT_ZONE_COLUMN,
        TEXT_ZONE_REGION,
        TEXT_ZONE_PARAGRAPH,
        TEXT_ZONE_LINE,
        TEXT_ZONE_WORD,
        TEXT_ZONE_CHARACTER
    ]

    def test_type(self):
        for zone in self.zones:
            assert_equal(type(zone), TextZoneType)
            assert_is_instance(zone, Symbol)

    def test_repr(self):
        assert_repr(TEXT_ZONE_PAGE, '<djvu.const.TextZoneType: page>')
        assert_repr(TEXT_ZONE_COLUMN, '<djvu.const.TextZoneType: column>')
        assert_repr(TEXT_ZONE_REGION, '<djvu.const.TextZoneType: region>')
        assert_repr(TEXT_ZONE_PARAGRAPH, '<djvu.const.TextZoneType: para>')
        assert_repr(TEXT_ZONE_LINE, '<djvu.const.TextZoneType: line>')
        assert_repr(TEXT_ZONE_WORD, '<djvu.const.TextZoneType: word>')
        assert_repr(TEXT_ZONE_CHARACTER, '<djvu.const.TextZoneType: char>')

    def test_identity(self):
        assert_is(TEXT_ZONE_PAGE, get_text_zone_type(Symbol('page')))
        assert_is(TEXT_ZONE_COLUMN, get_text_zone_type(Symbol('column')))
        assert_is(TEXT_ZONE_REGION, get_text_zone_type(Symbol('region')))
        assert_is(TEXT_ZONE_PARAGRAPH, get_text_zone_type(Symbol('para')))
        assert_is(TEXT_ZONE_LINE, get_text_zone_type(Symbol('line')))
        assert_is(TEXT_ZONE_WORD, get_text_zone_type(Symbol('word')))
        assert_is(TEXT_ZONE_CHARACTER, get_text_zone_type(Symbol('char')))

    def test_comparison1(self):
        assert_not_equal(TEXT_ZONE_PAGE, '')
        assert_not_equal(TEXT_ZONE_PAGE, 42)
        with assert_raises_str(TypeError, 'cannot compare text zone type with other object'):
            TEXT_ZONE_PAGE < 42
        with assert_raises_str(TypeError, 'cannot compare text zone type with other object'):
            TEXT_ZONE_PAGE <= 42
        with assert_raises_str(TypeError, 'cannot compare text zone type with other object'):
            TEXT_ZONE_PAGE > 42
        with assert_raises_str(TypeError, 'cannot compare text zone type with other object'):
            TEXT_ZONE_PAGE >= 42

    def test_comparison2(self):
        assert_equal(self.zones, sorted(self.zones, reverse=True))
        assert_equal(
            [[cmp(z1, z2) for z1 in self.zones] for z2 in self.zones], [
                [0, -1, -1, -1, -1, -1, -1],
                [+1, 0, -1, -1, -1, -1, -1],
                [+1, +1, 0, -1, -1, -1, -1],
                [+1, +1, +1, 0, -1, -1, -1],
                [+1, +1, +1, +1, 0, -1, -1],
                [+1, +1, +1, +1, +1, 0, -1],
                [+1, +1, +1, +1, +1, +1, 0],
            ]
        )

def test_wildcard_import():
    ns = wildcard_import('djvu.const')
    assert_list_equal(
        sorted(ns.keys()), [
            'ANNOTATION_ALIGN',
            'ANNOTATION_BACKGROUND',
            'ANNOTATION_MAPAREA',
            'ANNOTATION_METADATA',
            'ANNOTATION_MODE',
            'ANNOTATION_PRINTED_FOOTER',
            'ANNOTATION_PRINTED_HEADER',
            'ANNOTATION_ZOOM',
            'EMPTY_LIST',
            'EMPTY_OUTLINE',
            'MAPAREA_ARROW',
            'MAPAREA_BACKGROUND_COLOR',
            'MAPAREA_BORDER_ALWAYS_VISIBLE',
            'MAPAREA_BORDER_ETCHED_IN',
            'MAPAREA_BORDER_ETCHED_OUT',
            'MAPAREA_BORDER_NONE',
            'MAPAREA_BORDER_SHADOW_IN',
            'MAPAREA_BORDER_SHADOW_OUT',
            'MAPAREA_BORDER_SOLID_COLOR',
            'MAPAREA_BORDER_XOR',
            'MAPAREA_HIGHLIGHT_COLOR',
            'MAPAREA_LINE_COLOR',
            'MAPAREA_LINE_COLOR_DEFAULT',
            'MAPAREA_LINE_MIN_WIDTH',
            'MAPAREA_LINE_WIDTH',
            'MAPAREA_OPACITY',
            'MAPAREA_OPACITY_DEFAULT',
            'MAPAREA_OPACITY_MAX',
            'MAPAREA_OPACITY_MIN',
            'MAPAREA_PUSHPIN',
            'MAPAREA_SHADOW_BORDERS',
            'MAPAREA_SHADOW_BORDER_MAX_WIDTH',
            'MAPAREA_SHADOW_BORDER_MIN_WIDTH',
            'MAPAREA_SHAPE_LINE',
            'MAPAREA_SHAPE_OVAL',
            'MAPAREA_SHAPE_POLYGON',
            'MAPAREA_SHAPE_RECTANGLE',
            'MAPAREA_SHAPE_TEXT',
            'MAPAREA_TEXT_COLOR',
            'MAPAREA_TEXT_COLOR_DEFAULT',
            'MAPAREA_URI',
            'MAPAREA_URL',
            'METADATA_BIBTEX_KEYS',
            'METADATA_KEYS',
            'METADATA_PDFINFO_KEYS',
            'PRINTED_FOOTER_ALIGN_CENTER',
            'PRINTED_FOOTER_ALIGN_LEFT',
            'PRINTED_FOOTER_ALIGN_RIGHT',
            'PRINTER_HEADER_ALIGN_CENTER',
            'PRINTER_HEADER_ALIGN_LEFT',
            'PRINTER_HEADER_ALIGN_RIGHT',
            'TEXT_ZONE_CHARACTER',
            'TEXT_ZONE_COLUMN',
            'TEXT_ZONE_LINE',
            'TEXT_ZONE_PAGE',
            'TEXT_ZONE_PARAGRAPH',
            'TEXT_ZONE_REGION',
            'TEXT_ZONE_SEPARATORS',
            'TEXT_ZONE_WORD',
            'TextZoneType',
            'get_text_zone_type'
        ]
    )

# vim:ts=4 sts=4 sw=4 et



# Copyright © 2007-2015 Jakub Wilk <jwilk@jwilk.net>
#
# This file is part of python-djvulibre.
#
# python-djvulibre is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# python-djvulibre is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.

#cython: autotestdict=False

'''
DjVuLibre bindings: module for handling Lisp S-expressions
'''

cimport cython

include 'common.pxi'

cdef extern from 'libdjvu/miniexp.h':
    int cexpr_is_int 'miniexp_numberp'(cexpr_t sexp) nogil
    int cexpr_to_int 'miniexp_to_int'(cexpr_t sexp) nogil
    cexpr_t int_to_cexpr 'miniexp_number'(int n) nogil

    int cexpr_is_symbol 'miniexp_symbolp'(cexpr_t sexp) nogil
    char* cexpr_to_symbol 'miniexp_to_name'(cexpr_t sexp) nogil
    cexpr_t symbol_to_cexpr 'miniexp_symbol'(char* name) nogil

    cexpr_t cexpr_nil 'miniexp_nil'
    cexpr_t cexpr_dummy 'miniexp_dummy'
    int cexpr_is_list 'miniexp_listp'(cexpr_t exp) nogil
    int cexpr_is_nonempty_list 'miniexp_consp'(cexpr_t exp) nogil
    int cexpr_length 'miniexp_length'(cexpr_t exp) nogil
    cexpr_t cexpr_head 'miniexp_car'(cexpr_t exp) nogil
    cexpr_t cexpr_tail 'miniexp_cdr'(cexpr_t exp) nogil
    cexpr_t cexpr_nth 'miniexp_nth'(int n, cexpr_t exp) nogil
    cexpr_t pair_to_cexpr 'miniexp_cons'(cexpr_t head, cexpr_t tail) nogil
    cexpr_t cexpr_replace_head 'miniexp_rplaca'(cexpr_t exp, cexpr_t new_head) nogil
    cexpr_t cexpr_replace_tail 'miniexp_rplacd'(cexpr_t exp, cexpr_t new_tail) nogil
    cexpr_t cexpr_reverse_list 'miniexp_reverse'(cexpr_t exp) nogil

    int cexpr_is_str 'miniexp_stringp'(cexpr_t cexpr) nogil
    const char * cexpr_to_str 'miniexp_to_str'(cexpr_t cexpr) nogil
    cexpr_t str_to_cexpr 'miniexp_string'(const char *s) nogil
    cexpr_t cexpr_substr 'miniexp_substring'(const char *s, int n) nogil
    cexpr_t cexpr_concat 'miniexp_concat'(cexpr_t cexpr_list) nogil

    cexpr_t gc_lock 'minilisp_acquire_gc_lock'(cexpr_t cexpr) nogil
    cexpr_t gc_unlock 'minilisp_release_gc_lock'(cexpr_t cexpr) nogil

    cvar_t* cvar_new 'minivar_alloc'() nogil
    void cvar_free 'minivar_free'(cvar_t* v) nogil
    cexpr_t* cvar_ptr 'minivar_pointer'(cvar_t* v) nogil

    IF HAVE_MINIEXP_IO_T:
        ctypedef cexpr_io_s cexpr_io_t 'miniexp_io_t'
        struct cexpr_io_s 'miniexp_io_s':
            int (*puts 'fputs')(cexpr_io_t*, char*)
            int (*getc 'fgetc')(cexpr_io_t*)
            int (*ungetc)(cexpr_io_t*, int)
            void *data[4]
            int *p_flags
        void cexpr_io_init 'miniexp_io_init'(cexpr_io_t *cio)
        enum:
            cexpr_io_print7bits 'miniexp_io_print7bits'
        cexpr_t cexpr_read 'miniexp_read_r'(cexpr_io_t *cio)
        cexpr_t cexpr_print 'miniexp_prin_r'(cexpr_io_t *cio, cexpr_t cexpr)
        cexpr_t cexpr_printw 'miniexp_pprin_r'(cexpr_io_t *cio, cexpr_t cexpr, int width)
    ELSE:
        int io_7bit 'minilisp_print_7bits'
        int (*io_puts 'minilisp_puts')(char *s)
        int (*io_getc 'minilisp_getc')()
        int (*io_ungetc 'minilisp_ungetc')(int c)
        cexpr_t cexpr_read 'miniexp_read'()
        cexpr_t cexpr_print 'miniexp_prin'(cexpr_t cexpr)
        cexpr_t cexpr_printw 'miniexp_pprin'(cexpr_t cexpr, int width)

cdef extern from 'stdio.h':
    int EOF

cdef object sys
import sys

cdef object format_exc
from traceback import format_exc

cdef object StringIO
IF PY3K:
    from io import StringIO
ELSE:
    from cStringIO import StringIO

cdef object BytesIO
from io import BytesIO

cdef object weakref
import weakref

cdef object symbol_dict
symbol_dict = weakref.WeakValueDictionary()

cdef object codecs
import codecs

IF not HAVE_MINIEXP_IO_T:
    cdef Lock _myio_lock
    _myio_lock = allocate_lock()

cdef class _ExpressionIO:
    IF HAVE_MINIEXP_IO_T:
        cdef cexpr_io_t cio
        cdef int flags
    ELSE:
        cdef int (*backup_io_puts)(const char *s)
        cdef int (*backup_io_getc)()
        cdef int (*backup_io_ungetc)(int c)
        cdef int backup_io_7bit
    cdef object stdin 'stdin_fp'
    cdef object stdout 'stdout_fp'
    cdef int stdout_binary
    cdef object buffer
    cdef object exc

    _reentrant = HAVE_MINIEXP_IO_T

    def __init__(self, object stdin=None, object stdout=None, int escape_unicode=True):
        IF not HAVE_MINIEXP_IO_T:
            global io_7bit, io_puts, io_getc, io_ungetc
            global _myio
            with nogil: acquire_lock(_myio_lock, WAIT_LOCK)
            self.backup_io_7bit = io_7bit
            self.backup_io_puts = io_puts
            self.backup_io_getc = io_getc
            self.backup_io_ungetc = io_ungetc
        self.stdin = stdin
        self.stdout = stdout
        IF PY3K:
            self.stdout_binary = not hasattr(stdout, 'encoding')
        ELSE:
            # In Python 2, sys.stdout has the encoding attribute,
            # even though it accepts byte strings.
            # Let's only make a special-case for codecs.
            self.stdout_binary = not isinstance(stdout, codecs.StreamReaderWriter)
        self.buffer = []
        self.exc = None
        IF HAVE_MINIEXP_IO_T:
            cexpr_io_init(&self.cio)
            self.cio.data[0] = <void*>self
            self.cio.getc = _myio_getc
            self.cio.ungetc = _myio_ungetc
            self.cio.puts = _myio_puts
            if escape_unicode:
                self.flags = cexpr_io_print7bits
            else:
                self.flags = 0
            self.cio.p_flags = &self.flags
        ELSE:
            io_getc = _myio_getc
            io_ungetc = _myio_ungetc
            io_puts = _myio_puts
            io_7bit = escape_unicode
            _myio = self

    @cython.final
    cdef close(self):
        IF not HAVE_MINIEXP_IO_T:
            global io_7bit, io_puts, io_getc, io_ungetc
            global _myio
            _myio = None
        self.stdin = None
        self.stdout = None
        self.buffer = None
        IF not HAVE_MINIEXP_IO_T:
            io_7bit = self.backup_io_7bit
            io_puts = self.backup_io_puts
            io_getc = self.backup_io_getc
            io_ungetc = self.backup_io_ungetc
        try:
            if self.exc is not None:
                raise self.exc[0], self.exc[1], self.exc[2]
        finally:
            IF not HAVE_MINIEXP_IO_T:
                release_lock(_myio_lock)
            self.exc = None

    IF HAVE_MINIEXP_IO_T:

        @cython.final
        cdef cexpr_t read(self):
            return cexpr_read(&self.cio)

        @cython.final
        cdef cexpr_t print_(self, cexpr_t cexpr):
            return cexpr_print(&self.cio, cexpr)

        @cython.final
        cdef cexpr_t printw(self, cexpr_t cexpr, int width):
            return cexpr_printw(&self.cio, cexpr, width)

    ELSE:

        @cython.final
        cdef cexpr_t read(self):
            return cexpr_read()

        @cython.final
        cdef cexpr_t print_(self, cexpr_t cexpr):
            return cexpr_print(cexpr)

        @cython.final
        cdef cexpr_t printw(self, cexpr_t cexpr, int width):
            return cexpr_printw(cexpr, width)

IF HAVE_MINIEXP_IO_T:

    cdef int _myio_puts(cexpr_io_t* cio, const char *s):
        cdef _ExpressionIO io
        xio = <_ExpressionIO> cio.data[0]
        try:
            if xio.stdout_binary:
                xio.stdout.write(s)
            else:
                xio.stdout.write(decode_utf8(s))
        except:
            xio.exc = sys.exc_info()
            return EOF

    cdef int _myio_getc(cexpr_io_t* cio):
        cdef _ExpressionIO xio
        cdef int result
        xio = <_ExpressionIO> cio.data[0]
        if xio.buffer:
            return xio.buffer.pop()
        try:
            s = xio.stdin.read(1)
            if not s:
                return EOF
            if is_unicode(s):
                s = encode_utf8(s)
            IF PY3K:
                xio.buffer += reversed(s)
            ELSE:
                xio.buffer += map(ord, reversed(s))
            return xio.buffer.pop()
        except:
            xio.exc = sys.exc_info()
            return EOF

    cdef int _myio_ungetc(cexpr_io_t* cio, int c):
        cdef _ExpressionIO io
        xio = <_ExpressionIO> cio.data[0]
        list_append(xio.buffer, c)

ELSE:

    cdef _ExpressionIO _myio

    cdef int _myio_puts(const char *s):
        try:
            if _myio.stdout_binary:
                _myio.stdout.write(s)
            else:
                _myio.stdout.write(decode_utf8(s))
        except:
            _myio.exc = sys.exc_info()
            return EOF

    cdef int _myio_getc():
        cdef int result
        if _myio.buffer:
            return _myio.buffer.pop()
        try:
            s = _myio.stdin.read(1)
            if not s:
                return EOF
            if is_unicode(s):
                s = encode_utf8(s)
            IF PY3K:
                _myio.buffer += reversed(s)
            ELSE:
                _myio.buffer += map(ord, reversed(s))
            return _myio.buffer.pop()
        except:
            _myio.exc = sys.exc_info()
            return EOF

    cdef int _myio_ungetc(int c):
        list_append(_myio.buffer, c)

cdef object the_sentinel
the_sentinel = object()

cdef class _WrappedCExpr:

    def __cinit__(self, object sentinel):
        if sentinel is not the_sentinel:
            raise_instantiation_error(type(self))
        self.cvar = cvar_new()

    cdef cexpr_t cexpr(self):
        return cvar_ptr(self.cvar)[0]

    cdef object print_into(self, object stdout, object width, bint escape_unicode):
        cdef cexpr_t cexpr
        cdef _ExpressionIO xio
        if width is None:
            pass
        elif not is_int(width):
            raise TypeError('width must be an integer')
        elif width <= 0:
            raise ValueError('width <= 0')
        cexpr = self.cexpr()
        xio = _ExpressionIO(stdout=stdout, escape_unicode=escape_unicode)
        try:
            if width is None:
                xio.print_(cexpr)
            else:
                xio.printw(cexpr, width)
        finally:
            xio.close()

    cdef object as_string(self, object width, bint escape_unicode):
        stdout = StringIO()
        try:
            self.print_into(stdout, width, escape_unicode)
            return stdout.getvalue()
        finally:
            stdout.close()

    def __dealloc__(self):
        cvar_free(self.cvar)

cdef _WrappedCExpr wexpr(cexpr_t cexpr):
    cdef _WrappedCExpr wexpr
    wexpr = _WrappedCExpr(sentinel = the_sentinel)
    cvar_ptr(wexpr.cvar)[0] = cexpr
    return wexpr

cdef class _MissingCExpr(_WrappedCExpr):

    cdef object print_into(self, object stdout, object width, bint escape_unicode):
        raise NotImplementedError

    cdef object as_string(self, object width, bint escape_unicode):
        raise NotImplementedError

cdef _MissingCExpr wexpr_missing():
    return _MissingCExpr(the_sentinel)


cdef class BaseSymbol:

    cdef object __weakref__
    cdef object _bytes

    def __cinit__(self, bytes):
        cdef char *cbytes
        cbytes = bytes
        self._bytes = cbytes

    def __repr__(self):
        IF PY3K:
            try:
                string = self._bytes.decode('UTF-8')
            except UnicodeDecodeError:
                string = self._bytes
        ELSE:
            string = self._bytes
        return '{tp}({s!r})'.format(tp=get_type_name(_Symbol_), s=string)

    def __richcmp__(self, object other, int op):
        cdef BaseSymbol _self, _other
        if not typecheck(self, BaseSymbol) or not typecheck(other, BaseSymbol):
            return NotImplemented
        _self = self
        _other = other
        return richcmp(_self._bytes, _other._bytes, op)

    def __hash__(self):
        return hash(self._bytes)

    property bytes:
        def __get__(self):
            return self._bytes

    IF not PY3K:
        def __str__(self):
            return self._bytes

    IF PY3K:
        def __str__(self):
            return self._bytes.decode('UTF-8')
    ELSE:
        def __unicode__(self):
            return self._bytes.decode('UTF-8')

    def __reduce__(self):
        return (Symbol, (self._bytes,))

class Symbol(BaseSymbol):

    @staticmethod
    def __new__(cls, name):
        '''
        Symbol(name) -> a symbol
        '''
        self = None
        if is_unicode(name):
            name = encode_utf8(name)
        try:
            if cls is _Symbol_:
                self = symbol_dict[name]
        except KeyError:
            pass
        if self is None:
            if not is_bytes(name):
                name = str(name)
                IF PY3K:
                    name = encode_utf8(name)
            self = BaseSymbol.__new__(cls, name)
            if cls is _Symbol_:
                symbol_dict[name] = self
        return self

cdef object _Symbol_
_Symbol_ = Symbol

def _expression_from_string(s):
    '''
    Expression.from_string(s) -> an expression

    Read an expression from a string.
    '''
    if is_unicode(s):
        s = encode_utf8(s)
    stdin = BytesIO(s)
    try:
        return _Expression_.from_stream(stdin)
    finally:
        stdin.close()

class Expression(BaseExpression):

    '''
    Notes about the textual representation of S-expressions
    -------------------------------------------------------

    Special characters are:

    * the parenthesis '(' and ')',
    * the double quote '"',
    * the vertical bar '|'.

    Symbols are represented by their name. Vertical bars | can be used to
    delimit names that contain blanks, special characters, non printable
    characters, non-ASCII characters, or can be confused as a number.

    Numbers follow the syntax specified by the C function strtol() with
    base=0.

    Strings are delimited by double quotes. All C string escapes are
    recognized. Non-printable ASCII characters must be escaped.

    List are represented by an open parenthesis '(' followed by the space
    separated list elements, followed by a closing parenthesis ')'.

    When the cdr of the last pair is non zero, the closed parenthesis is
    preceded by a space, a dot '.', a space, and the textual representation
    of the cdr. (This is only partially supported by Python bindings.)

    '''

    @staticmethod
    def __new__(cls, value):
        '''
        Expression(value) -> an expression
        '''
        if typecheck(value, _Expression_) and (not typecheck(value, ListExpression) or not value):
            return value
        if is_int(value):
            return IntExpression(value)
        elif typecheck(value, _Symbol_):
            return SymbolExpression(value)
        elif is_unicode(value):
            return StringExpression(encode_utf8(value))
        elif is_bytes(value):
            if PY3K:
                return StringExpression(bytes(value))
            else:
                return StringExpression(str(value))
        else:
            return ListExpression(iter(value))

    @staticmethod
    def from_stream(stdin):
        '''
        Expression.from_stream(stream) -> an expression

        Read an expression from a stream.
        '''
        cdef _ExpressionIO xio
        try:
            xio = _ExpressionIO(stdin=stdin)
            try:
                return _c2py(xio.read())
            except InvalidExpression:
                raise ExpressionSyntaxError
        finally:
            xio.close()

    from_string = staticmethod(_expression_from_string)

cdef object _Expression_
_Expression_ = Expression

cdef object BaseExpression_richcmp(object left, object right, int op):
    if not typecheck(left, BaseExpression):
        return NotImplemented
    elif not typecheck(right, BaseExpression):
        return NotImplemented
    return richcmp(left.value, right.value, op)

cdef class BaseExpression:
    '''
    Don't use this class directly. Use the Expression class instead.
    '''

    cdef _WrappedCExpr wexpr

    def __cinit__(self, *args, **kwargs):
        self.wexpr = wexpr_missing()

    def print_into(self, stdout, width=None, escape_unicode=True):
        '''
        expr.print_into(file, width=None, escape_unicode=True) -> None

        Print the expression into the file.
        '''
        self.wexpr.print_into(stdout, width, escape_unicode)

    def as_string(self, width=None, escape_unicode=True):
        '''
        expr.as_string(width=None, escape_unicode=True) -> a string

        Return a string representation of the expression.
        '''
        return self.wexpr.as_string(width, escape_unicode)

    def __str__(self):
        return self.as_string()

    IF not PY3K:
        def __unicode__(self):
            return self.as_string().decode('UTF-8')

    property value:
        '''
        The "pythonic" value of the expression.
        Lisp lists as mapped to Python tuples.
        '''
        def __get__(self):
            return self._get_value()

    property lvalue:
        '''
        The "pythonic" value of the expression.
        Lisp lists as mapped to Python lists.
        '''
        def __get__(self):
            return self._get_lvalue()

    def _get_value(self):
        return self._get_lvalue()

    def _get_lvalue(self):
        raise NotImplementedError

    def __richcmp__(self, other, int op):
        return BaseExpression_richcmp(self, other, op)

    def __repr__(self):
        return '{tp}({expr!r})'.format(tp=get_type_name(_Expression_), expr=self.lvalue)

    def __copy__(self):
        # Most of S-expressions are immutable.
        # Mutable S-expressions should override this method.
        return self

    def __deepcopy__(self, memo):
        # Most of S-expressions are immutable.
        # Mutable S-expressions should override this method.
        return self

    def __reduce__(self):
        return (_expression_from_string, (self.as_string(),))

class IntExpression(_Expression_):

    '''
    IntExpression can represent any integer in range(-2 ** 29, 2 ** 29).

    To create objects of this class, use the Expression class constructor.
    '''

    @staticmethod
    def __new__(cls, value):
        '''
        IntExpression(n) -> an integer expression
        '''
        cdef BaseExpression self
        self = BaseExpression.__new__(cls)
        if typecheck(value, _WrappedCExpr):
            self.wexpr = value
        elif is_int(value):
            if -1 << 29 <= value < 1 << 29:
                self.wexpr = wexpr(int_to_cexpr(value))
            else:
                raise ValueError('value not in range(-2 ** 29, 2 ** 29)')
        else:
            raise TypeError('value must be an integer')
        return self

    IF PY3K:
        def __bool__(self):
            return bool(self.value)
    ELSE:
        def __nonzero__(self):
            return bool(self.value)

    def __int__(self):
        return self.value

    def __long__(self):
        return long(self.value)

    def __float__(self):
        return 0.0 + self.value

    def _get_lvalue(BaseExpression self not None):
        return cexpr_to_int(self.wexpr.cexpr())

    def __richcmp__(self, other, int op):
        return BaseExpression_richcmp(self, other, op)

    def __hash__(self):
        return hash(self.value)

class SymbolExpression(_Expression_):
    '''
    To create objects of this class, use the Expression class constructor.
    '''

    @staticmethod
    def __new__(cls, value):
        '''
        SymbolExpression(Symbol(s)) -> a symbol expression
        '''
        cdef BaseExpression self
        cdef BaseSymbol symbol
        self = BaseExpression.__new__(cls)
        if typecheck(value, _WrappedCExpr):
            self.wexpr = value
        elif typecheck(value, _Symbol_):
            symbol = value
            self.wexpr = wexpr(symbol_to_cexpr(symbol._bytes))
        else:
            raise TypeError('value must be a Symbol')
        return self

    def _get_lvalue(BaseExpression self not None):
        return _Symbol_(cexpr_to_symbol(self.wexpr.cexpr()))

    def __richcmp__(self, other, int op):
        return BaseExpression_richcmp(self, other, op)

    def __hash__(self):
        return hash(self.value)

class StringExpression(_Expression_):
    '''
    To create objects of this class, use the Expression class constructor.
    '''

    @staticmethod
    def __new__(cls, value):
        '''
        SymbolExpression(s) -> a string expression
        '''
        cdef BaseExpression self
        self = BaseExpression.__new__(cls)
        if typecheck(value, _WrappedCExpr):
            self.wexpr = value
        elif is_bytes(value):
            gc_lock(NULL)  # protect from collecting a just-created object
            try:
                self.wexpr = wexpr(str_to_cexpr(value))
            finally:
                gc_unlock(NULL)
        else:
            raise TypeError('value must be a byte string')
        return self

    @property
    def bytes(BaseExpression self not None):
        return cexpr_to_str(self.wexpr.cexpr())

    def _get_lvalue(BaseExpression self not None):
        cdef const char *bytes
        bytes = cexpr_to_str(self.wexpr.cexpr())
        IF PY3K:
            return decode_utf8(bytes)
        ELSE:
            return bytes

    IF PY3K:
        def __repr__(BaseExpression self not None):
            cdef const char *bytes
            bytes = cexpr_to_str(self.wexpr.cexpr())
            try:
                string = decode_utf8(bytes)
            except UnicodeDecodeError:
                string = bytes
            return '{tp}({s!r})'.format(tp=get_type_name(_Expression_), s=string)

    def __richcmp__(self, other, int op):
        return BaseExpression_richcmp(self, other, op)

    def __hash__(self):
        return hash(self.value)

class InvalidExpression(ValueError):
    pass

class ExpressionSyntaxError(Exception):
    '''
    Invalid expression syntax.
    '''
    pass

cdef _WrappedCExpr public_py2c(object o):
    cdef BaseExpression pyexpr
    pyexpr = _Expression_(o)
    if pyexpr is None:
        raise TypeError
    return pyexpr.wexpr

cdef object public_c2py(cexpr_t cexpr):
    return _c2py(cexpr)

cdef BaseExpression _c2py(cexpr_t cexpr):
    if cexpr == cexpr_dummy:
        raise InvalidExpression
    _wexpr = wexpr(cexpr)
    if cexpr_is_int(cexpr):
        result = IntExpression(_wexpr)
    elif cexpr_is_symbol(cexpr):
        result = SymbolExpression(_wexpr)
    elif cexpr_is_list(cexpr):
        result = ListExpression(_wexpr)
    elif cexpr_is_str(cexpr):
        result = StringExpression(_wexpr)
    else:
        raise InvalidExpression
    return result

cdef _WrappedCExpr _build_list_cexpr(object items):
    cdef cexpr_t cexpr
    cdef BaseExpression citem
    gc_lock(NULL)  # protect from collecting a just-created object
    try:
        cexpr = cexpr_nil
        for item in items:
            if typecheck(item, BaseExpression):
                citem = item
            else:
                citem = _Expression_(item)
            if citem is None:
                raise TypeError
            cexpr = pair_to_cexpr(citem.wexpr.cexpr(), cexpr)
        cexpr = cexpr_reverse_list(cexpr)
        return wexpr(cexpr)
    finally:
        gc_unlock(NULL)


class ListExpression(_Expression_):
    '''
    To create objects of this class, use the Expression class constructor.
    '''

    @staticmethod
    def __new__(cls, items):
        '''
        ListExpression(iterable) -> a list expression
        '''
        cdef BaseExpression self
        self = BaseExpression.__new__(cls)
        if typecheck(items, _WrappedCExpr):
            self.wexpr = items
        else:
            self.wexpr = _build_list_cexpr(items)
        return self

    IF PY3K:
        def __bool__(BaseExpression self not None):
            return self.wexpr.cexpr() != cexpr_nil
    ELSE:
        def __nonzero__(BaseExpression self not None):
            return self.wexpr.cexpr() != cexpr_nil

    def __len__(BaseExpression self not None):
        cdef cexpr_t cexpr
        cdef int n
        cexpr = self.wexpr.cexpr()
        n = 0
        while cexpr != cexpr_nil:
            cexpr = cexpr_tail(cexpr)
            n = n + 1
        return n

    def __getitem__(BaseExpression self not None, key):
        cdef cexpr_t cexpr
        cdef int n
        cexpr = self.wexpr.cexpr()
        if is_int(key):
            n = key
            if n < 0:
                n = n + len(self)
            if n < 0:
                raise IndexError('list index of out range')
            while 1:
                if cexpr == cexpr_nil:
                    raise IndexError('list index of out range')
                if n > 0:
                    n = n - 1
                    cexpr = cexpr_tail(cexpr)
                else:
                    cexpr = cexpr_head(cexpr)
                    break
        elif is_slice(key):
            if (is_int(key.start) or key.start is None) and key.stop is None and key.step is None:
                n = key.start or 0
                if n < 0:
                    n = n + len(self)
                while n > 0 and cexpr != cexpr_nil:
                    cexpr = cexpr_tail(cexpr)
                    n = n - 1
            else:
                raise NotImplementedError('only [n:] slices are supported')
        else:
            raise TypeError('key must be an integer or a slice')
        return _c2py(cexpr)

    def __setitem__(BaseExpression self not None, key, value):
        cdef cexpr_t cexpr
        cdef cexpr_t prev_cexpr
        cdef cexpr_t new_cexpr
        cdef int n
        cdef BaseExpression pyexpr
        cexpr = self.wexpr.cexpr()
        pyexpr = _Expression_(value)
        new_cexpr = pyexpr.wexpr.cexpr()
        if is_int(key):
            n = key
            if n < 0:
                n = n + len(self)
            if n < 0:
                raise IndexError('list index of out range')
            while 1:
                if cexpr == cexpr_nil:
                    raise IndexError('list index of out range')
                if n > 0:
                    n = n - 1
                    cexpr = cexpr_tail(cexpr)
                else:
                    cexpr_replace_head(cexpr, new_cexpr)
                    break
        elif is_slice(key):
            if not cexpr_is_list(new_cexpr):
                raise TypeError('can only assign a list expression')
            if (is_int(key.start) or key.start is None) and key.stop is None and key.step is None:
                n = key.start or 0
                if n < 0:
                    n = n + len(self)
                prev_cexpr = cexpr_nil
                while n > 0 and cexpr != cexpr_nil:
                    prev_cexpr = cexpr
                    cexpr = cexpr_tail(cexpr)
                    n = n - 1
                if prev_cexpr == cexpr_nil:
                    self.wexpr = wexpr(new_cexpr)
                else:
                    cexpr_replace_tail(prev_cexpr, new_cexpr)
            else:
                raise NotImplementedError('only [n:] slices are supported')
        else:
            raise TypeError('key must be an integer or a slice')

    def __delitem__(BaseExpression self not None, key):
        if is_int(key):
            self.pop(key)
        elif is_slice(key):
            self[key] = ()
        else:
            raise TypeError('key must be an integer or a slice')

    def extend(self, iterable):
        iter(iterable)
        self[len(self):] = iterable

    def __iadd__(self, iterable):
        iter(iterable)
        self[len(self):] = iterable
        return self

    def insert(BaseExpression self not None, long index, item):
        cdef cexpr_t cexpr, new_cexpr
        cdef BaseExpression citem
        cexpr = self.wexpr.cexpr()
        if index < 0:
            index += len(self)
        if index < 0:
            index = 0
        if typecheck(item, BaseExpression):
            citem = item
        else:
            citem = _Expression_(item)
        if citem is None:
            raise TypeError
        if index == 0 or cexpr == cexpr_nil:
            gc_lock(NULL)  # protect from collecting a just-created object
            try:
                new_cexpr = pair_to_cexpr(citem.wexpr.cexpr(), cexpr)
                self.wexpr = wexpr(new_cexpr)
            finally:
                gc_unlock(NULL)
            return
        while 1:
            assert cexpr != cexpr_nil
            if index > 1 and cexpr_tail(cexpr) != cexpr_nil:
                index = index - 1
                cexpr = cexpr_tail(cexpr)
            else:
                gc_lock(NULL)  # protect from collecting a just-created object
                try:
                    new_cexpr = pair_to_cexpr(citem.wexpr.cexpr(), cexpr_tail(cexpr))
                    cexpr_replace_tail(cexpr, new_cexpr)
                finally:
                    gc_unlock(NULL)
                break

    def append(BaseExpression self not None, item):
        return self.insert(len(self), item)

    def reverse(BaseExpression self not None):
        cdef cexpr_t cexpr, new_cexpr
        gc_lock(NULL)  # protect from collecting a just-created object
        try:
            new_cexpr = cexpr_reverse_list(self.wexpr.cexpr())
            self.wexpr = wexpr(new_cexpr)
        finally:
            gc_unlock(NULL)

    def pop(BaseExpression self not None, long index=-1):
        cdef cexpr_t cexpr, citem
        cexpr = self.wexpr.cexpr()
        if cexpr == cexpr_nil:
            raise IndexError('pop from empty list')
        if index < 0:
            index += len(self)
        if index < 0:
            raise IndexError('pop index of out range')
        if index == 0:
            result = _c2py(cexpr_head(cexpr))
            self.wexpr = wexpr(cexpr_tail(cexpr))
            return result
        while cexpr_tail(cexpr) != cexpr_nil:
            if index > 1:
                index = index - 1
                cexpr = cexpr_tail(cexpr)
            else:
                result = _c2py(cexpr_head(cexpr_tail(cexpr)))
                cexpr_replace_tail(cexpr, cexpr_tail(cexpr_tail(cexpr)))
                return result
        raise IndexError('pop index of out range')

    def remove(BaseExpression self not None, item):
        cdef cexpr_t cexpr
        cdef BaseExpression citem
        cexpr = self.wexpr.cexpr()
        if cexpr == cexpr_nil:
            raise IndexError('item not in list')
        if _c2py(cexpr_head(cexpr)) == item:
            self.wexpr = wexpr(cexpr_tail(cexpr))
            return
        while 1:
            assert cexpr != cexpr_nil
            if cexpr_tail(cexpr) == cexpr_nil:
                raise IndexError('item not in list')
            if _c2py(cexpr_head(cexpr_tail(cexpr))) == item:
                cexpr_replace_tail(cexpr, cexpr_tail(cexpr_tail(cexpr)))
                return
            cexpr = cexpr_tail(cexpr)

    def index(self, value):
        # TODO: optimize
        for i, v in enumerate(self):
            if v == value:
                return i
        raise ValueError('value not in list')

    def count(self, value):
        # TODO: optimize
        cdef long counter = 0
        for v in self:
            if v == value:
                counter += 1
        return counter

    def __iter__(self):
        return _ListExpressionIterator(self)

    __hash__ = None

    def _get_value(BaseExpression self not None):
        cdef cexpr_t current
        current = self.wexpr.cexpr()
        result = []
        while current != cexpr_nil:
            list_append(result, _c2py(cexpr_head(current))._get_value())
            current = cexpr_tail(current)
        return tuple(result)

    def _get_lvalue(BaseExpression self not None):
        cdef cexpr_t current
        current = self.wexpr.cexpr()
        result = []
        while current != cexpr_nil:
            list_append(result, _c2py(cexpr_head(current))._get_lvalue())
            current = cexpr_tail(current)
        return result

    def __copy__(self):
        return _Expression_(self)

    def __deepcopy__(self, memo):
        return _Expression_(self._get_value())

import collections
collections.MutableSequence.register(ListExpression)
del collections

cdef class _ListExpressionIterator:

    cdef BaseExpression expression
    cdef cexpr_t cexpr

    def __cinit__(self, BaseExpression expression not None):
        self.expression = expression
        self.cexpr = expression.wexpr.cexpr()

    def __next__(self):
        cdef cexpr_t cexpr
        cexpr = self.cexpr
        if cexpr == cexpr_nil:
            raise StopIteration
        self.cexpr = cexpr_tail(cexpr)
        cexpr = cexpr_head(cexpr)
        return _c2py(cexpr)

    def __iter__(self):
        return self


__all__ = ('Symbol', 'Expression', 'IntExpression', 'SymbolExpression', 'StringExpression', 'ListExpression', 'InvalidExpression', 'ExpressionSyntaxError')
__author__ = 'Jakub Wilk <jwilk@jwilk.net>'
IF PY3K:
    __version__ = decode_utf8(PYTHON_DJVULIBRE_VERSION)
ELSE:
    __version__ = str(PYTHON_DJVULIBRE_VERSION)

# vim:ts=4 sts=4 sw=4 et ft=pyrex



# encoding=UTF-8

# Copyright © 2007-2015 Jakub Wilk <jwilk@jwilk.net>
#
# This file is part of python-djvulibre.
#
# python-djvulibre is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# python-djvulibre is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.

from __future__ import division

if __name__ != '__main__':
    raise ImportError('This module is not intended for import')

import djvu.sexpr
import os

proc_status = '/proc/{pid}/status'.format(pid=os.getpid())
scale = dict(kB=1024)

def mem_info(key='VmSize'):
    try:
        file = open(proc_status)
        for line in file:
            if line.startswith('{key}:'.format(key=key)):
                _, value, unit = line.split(None, 3)
                return int(value) * scale[unit]
    finally:
        file.close()

try:
    range = xrange
except NameError:
    pass

step = 1 << 17
while True:
    mb = mem_info() / (1 << 20)
    print('{mb:.2f}M'.format(mb=mb))
    [djvu.sexpr.Expression(4) for i in range(step)]

# vim:ts=4 sts=4 sw=4 et



# encoding=UTF-8

# Copyright © 2007-2016 Jakub Wilk <jwilk@jwilk.net>
#
# This file is part of python-djvulibre.
#
# python-djvulibre is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# python-djvulibre is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.

import codecs
import collections
import copy
import errno
import io
import os
import shutil
import sys
import tempfile

import pickle
try:
    import cPickle as cpickle
except ImportError:
    cpickle = None

from djvu.sexpr import (
    Expression,
    ExpressionSyntaxError,
    Symbol,
    _ExpressionIO,
    __version__,
)

from tools import (
    SkipTest,
    assert_equal,
    assert_false,
    assert_in,
    assert_is,
    assert_is_instance,
    assert_less,
    assert_list_equal,
    assert_not_equal,
    assert_not_in,
    assert_raises,
    assert_raises_str,
    assert_repr,
    wildcard_import,
    # Python 2/3 compat:
    StringIO,
    b,
    long,
    py3k,
    u,
    unicode,
)

def assert_pickle_equal(obj):
    for pickle_module in pickle, cpickle:
        if pickle_module is None:
            continue
        for protocol in range(pickle.HIGHEST_PROTOCOL + 1):
            pickled_obj = pickle_module.dumps(obj, protocol=protocol)
            repickled_obj = pickle_module.loads(pickled_obj)
            assert_equal(obj, repickled_obj)

class test_int_expressions():

    def t(self, n, x=None):
        if x is None:
            x = Expression(n)
        assert_is(x, Expression(x))
        # __repr__():
        assert_repr(x, 'Expression({n})'.format(n=int(n)))
        # value:
        v = x.value
        assert_equal(type(v), int)
        assert_equal(v, n)
        # lvalue:
        v = x.lvalue
        assert_equal(type(v), int)
        assert_equal(v, n)
        # __int__():
        i = int(x)
        assert_equal(type(i), int)
        assert_equal(i, n)
        # __long__():
        i = long(x)
        assert_equal(type(i), long)
        assert_equal(i, n)
        # __float__():
        i = float(x)
        assert_equal(type(i), float)
        assert_equal(i, n)
        # __str__():
        s = str(x)
        assert_equal(s, str(n))
        # __unicode__():
        s = unicode(x)
        assert_equal(s, str(n))
        # __eq__(), __ne__():
        assert_equal(x, Expression(n))
        assert_not_equal(x, n)
        assert_not_equal(x, Expression(n + 37))
        # __hash__():
        assert_equal(hash(x), n)
        # __bool__() / __nonzero__():
        obj = object()
        if n:
            assert_is(x and obj, obj)
            assert_is(x or obj, x)
        else:
            assert_is(x and obj, x)
            assert_is(x or obj, obj)
        # pickle:
        assert_pickle_equal(x)

    def test_int(self):
        self.t(42)

    def test_parse(self):
        self.t(42, Expression.from_string('42'))

    def test_unpickle(self):
        # pickle as generated by python-djvulibre 0.3.3:
        p = b("cdjvu.sexpr\n_expression_from_string\np0\n(S'42'\np1\ntp2\nRp3\n.")
        x = pickle.loads(p)
        self.t(42, x)

    def test_0(self):
        self.t(0)

    def test_long(self):
        self.t(long(42))

    def test_limits(self):
        assert_equal(Expression((1 << 29) - 1).value, (1 << 29) - 1)
        assert_equal(Expression(-1 << 29).value, -1 << 29)
        with assert_raises_str(ValueError, 'value not in range(-2 ** 29, 2 ** 29)'):
            Expression(1 << 29)
        with assert_raises_str(ValueError, 'value not in range(-2 ** 29, 2 ** 29)'):
            Expression((-1 << 29) - 1)

class test_float_expressions():

    # TODO: float expressions are not implemented yet

    def test_parse(self):
        with assert_raises(ExpressionSyntaxError):
            x = Expression.from_string('3.14')
            if isinstance(x.value, Symbol):
                raise ExpressionSyntaxError

class test_symbols():

    def t(self, name, sname=None):
        if sname is None:
            sname = name
        if py3k:
            [uname, bname] = [sname, sname.encode('UTF-8')]
        else:
            [uname, bname] = [sname.decode('UTF-8'), sname]
        symbol = Symbol(name)
        assert_equal(type(symbol), Symbol)
        assert_equal(symbol, Symbol(name))
        assert_is(symbol, Symbol(name))
        assert_equal(str(symbol), sname)
        assert_equal(unicode(symbol), uname)
        assert_not_equal(symbol, bname)
        assert_not_equal(symbol, uname)
        assert_equal(hash(symbol), hash(bname))
        assert_pickle_equal(symbol)
        return symbol

    def test_ascii(self):
        self.t('eggs')

    def test_nonascii(self):
        x = self.t(b('ветчина'), 'ветчина')
        y = self.t(u('ветчина'), 'ветчина')
        assert x is y

    def test_inequality(self):
        assert_less(
            Symbol('eggs'),
            Symbol('ham'),
        )

class test_symbol_expressions():

    def t(self, name, sname):
        if sname is None:
            sname = name
        if py3k:
            [uname, bname] = [sname, sname.encode('UTF-8')]
        else:
            [uname, bname] = [sname.decode('UTF-8'), sname]
        sym = Symbol(name)
        x = Expression(sym)
        assert_is(x, Expression(x))
        # __repr__(x)
        assert_repr(x, 'Expression({sym!r})'.format(sym=sym))
        # value:
        v = x.value
        assert_equal(type(v), Symbol)
        assert_equal(v, sym)
        # lvalue:
        v = x.lvalue
        assert_equal(type(v), Symbol)
        assert_equal(v, sym)
        # __str__():
        assert_equal(str(x), sname)
        assert_repr(x, repr(Expression.from_string(sname)))
        # __unicode__():
        assert_equal(unicode(x), uname)
        assert_repr(x, repr(Expression.from_string(uname)))
        # __eq__(), __ne__():
        assert_equal(x, Expression(sym))
        assert_not_equal(x, Expression(name))
        assert_not_equal(x, sym)
        # __hash__():
        assert_equal(
            hash(x),
            hash(bname.strip(b('|')))
        )
        # pickle:
        assert_pickle_equal(x)
        return x

    def test_ascii(self):
        self.t('eggs', 'eggs')

    def test_nonascii(self):
        x = self.t(b('ветчина'), '|ветчина|')
        y = self.t(u('ветчина'), '|ветчина|')
        assert_equal(x, y)
        assert_equal(hash(x), hash(y))

def test_string_expressions():
    x = Expression('eggs')
    assert_repr(x, "Expression('eggs')")
    assert_is(x, Expression(x))
    assert_equal(x.value, 'eggs')
    assert_equal(x.lvalue, 'eggs')
    assert_equal(str(x), '"eggs"')
    assert_repr(x, repr(Expression.from_string(str(x))))
    assert_equal(x, Expression('eggs'))
    assert_not_equal(x, Expression(Symbol('eggs')))
    assert_not_equal(x, 'eggs')
    assert_equal(hash(x), hash('eggs'))
    assert_pickle_equal(x)

class test_unicode_expressions():

    def test1(self):
        x = Expression(u('eggs'))
        assert_repr(x, "Expression('eggs')")
        assert_is(x, Expression(x))

    def test2(self):
        x = Expression(u('żółw'))
        if py3k:
            assert_repr(x, "Expression('żółw')")
        else:
            assert_repr(x, r"Expression('\xc5\xbc\xc3\xb3\xc5\x82w')")

class test_list_expressions():

    def test1(self):
        x = Expression(())
        assert_repr(x, "Expression([])")
        y = Expression(x)
        assert_is(x, y)
        assert_equal(x.value, ())
        assert_equal(x.lvalue, [])
        assert_equal(len(x), 0)
        assert_equal(bool(x), False)
        assert_equal(list(x), [])

    def test2(self):
        x = Expression([[1, 2], 3, [4, 5, Symbol('baz')], ['quux']])
        assert_repr(x, "Expression([[1, 2], 3, [4, 5, Symbol('baz')], ['quux']])")
        y = Expression(x)
        assert_repr(y, repr(x))
        assert_false(x is y)
        assert_equal(x.value, ((1, 2), 3, (4, 5, Symbol('baz')), ('quux',)))
        assert_equal(x.lvalue, [[1, 2], 3, [4, 5, Symbol('baz')], ['quux']])
        assert_equal(str(x), '((1 2) 3 (4 5 baz) ("quux"))')
        assert_repr(x, repr(Expression.from_string(str(x))))
        assert_equal(len(x), 4)
        assert_equal(bool(x), True)
        assert_equal(tuple(x), (Expression((1, 2)), Expression(3), Expression((4, 5, Symbol('baz'))), Expression(('quux',))))
        with assert_raises_str(TypeError, 'key must be an integer or a slice'):
            x[object()]
        assert_equal(x[1], Expression(3))
        assert_equal(x[-1][0], Expression('quux'))
        with assert_raises_str(IndexError, 'list index of out range'):
            x[6]
        with assert_raises_str(IndexError, 'list index of out range'):
            x[-6]
        assert_equal(x[:].value, x.value)
        assert_equal(x[:].lvalue, x.lvalue)
        assert_repr(x[1:], "Expression([3, [4, 5, Symbol('baz')], ['quux']])")
        assert_repr(x[-2:], "Expression([[4, 5, Symbol('baz')], ['quux']])")
        x[-2:] = 4, 5, 6
        assert_repr(x, 'Expression([[1, 2], 3, 4, 5, 6])')
        x[0] = 2
        assert_repr(x, 'Expression([2, 3, 4, 5, 6])')
        x[:] = (1, 3, 5)
        assert_repr(x, 'Expression([1, 3, 5])')
        x[3:] = 7,
        assert_repr(x, 'Expression([1, 3, 5, 7])')
        with assert_raises_str(NotImplementedError, 'only [n:] slices are supported'):
            x[object():]
        with assert_raises_str(NotImplementedError, 'only [n:] slices are supported'):
            x[:2]
        with assert_raises_str(NotImplementedError, 'only [n:] slices are supported'):
            x[object():] = []
        with assert_raises_str(NotImplementedError, 'only [n:] slices are supported'):
            x[:2] = []
        with assert_raises_str(TypeError, 'can only assign a list expression'):
            x[:] = 0
        assert_equal(x, Expression((1, 3, 5, 7)))
        assert_not_equal(x, Expression((2, 4, 6)))
        assert_not_equal(x, (1, 3, 5, 7))
        with assert_raises_str(TypeError, "unhashable type: 'ListExpression'"):
            hash(x)

    def test_insert(self):
        lst = []
        expr = Expression(())
        for pos in [-8, 4, 6, -5, -7, 5, 7, 2, -3, 8, 10, -2, 1, -9, -10, -4, -6, 0, 9, 3, -1]:
            lst.insert(pos, pos)
            assert_is(expr.insert(pos, pos), None)
            assert_equal(expr, Expression(lst))
            assert_equal(expr.lvalue, lst)

    def test_append(self):
        expr = Expression(())
        for i in range(10):
            assert_is(expr.append(i), None)
            assert_equal(expr, Expression(range(i + 1)))
            assert_equal(expr.lvalue, list(range(i + 1)))

    def test_extend(self):
        lst = []
        expr = Expression(())
        for ext in [1], [], [2, 3]:
            lst.extend(ext)
            expr.extend(ext)
            assert_equal(expr, Expression(lst))
            assert_equal(expr.lvalue, lst)
        with assert_raises_str(TypeError, "'int' object is not iterable"):
            expr.extend(0)

    def test_inplace_add(self):
        lst = []
        expr0 = expr = Expression(())
        for ext in [], [1], [], [2, 3]:
            lst += ext
            expr += ext
            assert_equal(expr, Expression(lst))
            assert_equal(expr.lvalue, lst)
        assert_is(expr, expr0)
        with assert_raises_str(TypeError, "'int' object is not iterable"):
            expr += 0

    def test_pop(self):
        expr = Expression([0, 1, 2, 3, 4, 5, 6])
        assert_equal(expr.pop(0), Expression(0))
        assert_equal(expr, Expression([1, 2, 3, 4, 5, 6]))
        with assert_raises_str(IndexError, 'pop index of out range'):
            expr.pop(6)
        assert_equal(expr.pop(5), Expression(6))
        assert_equal(expr, Expression([1, 2, 3, 4, 5]))
        assert_equal(expr.pop(-1), Expression(5))
        assert_equal(expr, Expression([1, 2, 3, 4]))
        assert_equal(expr.pop(-2), Expression(3))
        assert_equal(expr, Expression([1, 2, 4]))
        assert_equal(expr.pop(1), Expression(2))
        assert_equal(expr, Expression([1, 4]))
        expr.pop()
        expr.pop()
        with assert_raises_str(IndexError, 'pop from empty list'):
            expr.pop()
        for i in range(-2, 3):
            with assert_raises_str(IndexError, 'pop from empty list'):
                expr.pop(i)

    def test_delitem(self):
        expr = Expression([0, 1, 2, 3, 4, 5, 6])
        del expr[0]
        assert_equal(expr, Expression([1, 2, 3, 4, 5, 6]))
        with assert_raises_str(IndexError, 'pop index of out range'):
            expr.pop(6)
        del expr[5]
        assert_equal(expr, Expression([1, 2, 3, 4, 5]))
        del expr[-1]
        assert_equal(expr, Expression([1, 2, 3, 4]))
        del expr[-2]
        assert_equal(expr, Expression([1, 2, 4]))
        del expr[1]
        assert_equal(expr, Expression([1, 4]))
        del expr[1:]
        assert_equal(expr, Expression([1]))
        del expr[:]
        assert_equal(expr, Expression([]))
        for i in range(-2, 3):
            with assert_raises_str(IndexError, 'pop from empty list'):
                del expr[i]

    def test_remove(self):
        expr = Expression([0, 1, 2, 3, 4, 5, 6])
        expr.remove(Expression(0))
        assert_equal(expr, Expression([1, 2, 3, 4, 5, 6]))
        with assert_raises_str(IndexError, 'item not in list'):
            expr.remove(Expression(0))
        expr.remove(Expression(6))
        assert_equal(expr, Expression([1, 2, 3, 4, 5]))
        expr.remove(Expression(5))
        assert_equal(expr, Expression([1, 2, 3, 4]))
        expr.remove(Expression(3))
        assert_equal(expr, Expression([1, 2, 4]))
        expr.remove(Expression(2))
        assert_equal(expr, Expression([1, 4]))
        expr.remove(Expression(4))
        expr.remove(Expression(1))
        with assert_raises_str(IndexError, 'item not in list'):
            expr.remove(Expression(-1))

    def test_contains(self):
        expr = Expression(())
        assert_not_in(Expression(42), expr)
        lst = (1, 2, 3)
        expr = Expression(lst)
        for x in lst:
            assert_not_in(x, expr)
            assert_in(Expression(x), expr)
        assert_not_in(Expression(max(lst) + 1), expr)

    def test_index(self):
        expr = Expression(())
        with assert_raises_str(ValueError, 'value not in list'):
            expr.index(Expression(42))
        lst = [1, 2, 3]
        expr = Expression(lst)
        for x in lst:
            i = lst.index(x)
            j = expr.index(Expression(x))
            assert_equal(i, j)
        with assert_raises_str(ValueError, 'value not in list'):
            expr.index(Expression(max(lst) + 1))

    def test_count(self):
        lst = [1, 2, 2, 3, 2]
        expr = Expression(lst)
        for x in lst + [max(lst) + 1]:
            i = lst.count(x)
            j = expr.count(Expression(x))
            assert_equal(i, j)

    def test_reverse(self):
        for lst in (), (1, 2, 3):
            expr = Expression(lst)
            assert_equal(
                Expression(reversed(expr)),
                Expression(reversed(lst))
            )
            assert_equal(
                Expression(reversed(expr)).value,
                tuple(reversed(lst))
            )
            assert_is(expr.reverse(), None)
            assert_equal(
                expr,
                Expression(reversed(lst))
            )
            assert_equal(
                expr.value,
                tuple(reversed(lst))
            )

    def test_copy1(self):
        x = Expression([1, [2], 3])
        y = Expression(x)
        x[1][0] = 0
        assert_repr(x, 'Expression([1, [0], 3])')
        assert_repr(y, 'Expression([1, [0], 3])')
        x[1] = 0
        assert_repr(x, 'Expression([1, 0, 3])')
        assert_repr(y, 'Expression([1, [0], 3])')

    def test_copy2(self):
        x = Expression([1, [2], 3])
        y = copy.copy(x)
        x[1][0] = 0
        assert_repr(x, 'Expression([1, [0], 3])')
        assert_repr(y, 'Expression([1, [0], 3])')
        x[1] = 0
        assert_repr(x, 'Expression([1, 0, 3])')
        assert_repr(y, 'Expression([1, [0], 3])')

    def test_copy3(self):
        x = Expression([1, [2], 3])
        y = copy.deepcopy(x)
        x[1][0] = 0
        assert_repr(x, 'Expression([1, [0], 3])')
        assert_repr(y, 'Expression([1, [2], 3])')
        x[1] = 0
        assert_repr(x, 'Expression([1, 0, 3])')
        assert_repr(y, 'Expression([1, [2], 3])')

    def test_abc(self):
        x = Expression(())
        assert_is_instance(x, collections.MutableSequence)
        assert_is_instance(iter(x), collections.Iterator)

    def test_pickle(self):
        for lst in (), (1, 2, 3), (1, (2, 3)):
            x = Expression(lst)
            assert_pickle_equal(x)

class test_expression_parser():

    def test_badstring(self):
        with assert_raises(ExpressionSyntaxError):
            Expression.from_string('(1')

    def test_attr_from_file(self):
        assert_is(getattr(Expression, 'from_file', None), None)

    def test_bad_io(self):
        with assert_raises_str(AttributeError, "'int' object has no attribute 'read'"):
            Expression.from_stream(42)

    def test_bad_file_io(self):
        if os.name == 'nt':
            raise SkipTest('not implemented on Windows')
        path = '/proc/self/mem'
        try:
            os.stat(path)
        except OSError as exc:
            raise SkipTest('{exc.filename}: {exc.strerror}'.format(exc=exc))
        with open('/proc/self/mem') as fp:
            with assert_raises(IOError) as ecm:
                Expression.from_stream(fp)
        assert_in(
            ecm.exception.errno,
            (errno.EIO, errno.EFAULT)
        )

    if py3k:
        def test_bad_unicode_io(self):
            fp = StringIO(chr(0xD800))
            with assert_raises(UnicodeEncodeError):
                Expression.from_stream(fp)

class test_expression_parser_ascii():

    expr = '(eggs) (ham)'
    repr = ["Expression([Symbol('eggs')])", "Expression([Symbol('ham')])"]

    def t(self, fp):
        def read():
            return Expression.from_stream(fp)
        x = read()
        assert_repr(x, self.repr[0])
        x = read()
        assert_repr(x, self.repr[1])
        with assert_raises(ExpressionSyntaxError):
            x = read()

    def test_stringio(self):
        fp = StringIO(self.expr)
        self.t(fp)

    def test_bytesio(self):
        fp = io.BytesIO(b(self.expr))
        self.t(fp)

    def test_file_io_text(self):
        if py3k:
            fp = tempfile.TemporaryFile(mode='w+t', encoding='UTF-16-LE')
        else:
            fp = tempfile.TemporaryFile(mode='w+t')
        with fp:
            fp.write(self.expr)
            fp.flush()
            fp.seek(0)
            self.t(fp)

    def test_codecs_io(self):
        tmpdir = tempfile.mkdtemp()
        try:
            path = os.path.join(tmpdir, 'tmp')
            with codecs.open(path, mode='w+', encoding='UTF-16-LE') as fp:
                fp.write(u(self.expr))
                fp.seek(0)
                self.t(fp)
        finally:
            shutil.rmtree(tmpdir)

    def test_file_io_binary(self):
        with tempfile.TemporaryFile(mode='w+b') as fp:
            fp.write(b(self.expr))
            fp.flush()
            fp.seek(0)
            self.t(fp)

class test_expression_parser_nonascii(test_expression_parser_ascii):

    expr = '"jeż" "żółw"'
    repr = [r"Expression('jeż')", r"Expression('żółw')"]
    if not py3k:
        repr = [s.decode('ISO-8859-1').encode('ASCII', 'backslashreplace') for s in repr]

class test_expression_writer():

    def test_bad_io(self):
        expr = Expression(23)
        with assert_raises_str(AttributeError, "'int' object has no attribute 'write'"):
            expr.print_into(42)

    def test_bad_file_io(self):
        ecm = None
        path = '/dev/full'
        try:
            os.stat(path)
        except OSError as exc:
            raise SkipTest('{exc.filename}: {exc.strerror}'.format(exc=exc))
        fp = open(path, 'w', buffering=2)
        expr = Expression(23)
        try:
            with assert_raises(IOError) as ecm:
                for i in range(10000):
                    expr.print_into(fp)
        finally:
            try:
                fp.close()
            except IOError:
                if ecm is None:
                    raise
        assert_equal(ecm.exception.errno, errno.ENOSPC)

    def test_reentrant(self):
        if not _ExpressionIO._reentrant:
            raise SkipTest('this test requires DjVuLibre >= 3.5.26')
        class File(object):
            def write(self, s):
                expr.as_string()
        expr = Expression(23)
        fp = File()
        expr.print_into(fp)

    def test_escape_unicode_type(self):
        expr = Expression(23)
        fp = StringIO()
        for v in True, False, 1, 0, 'yes', '':
            expr.print_into(fp, escape_unicode=v)
            expr.as_string(escape_unicode=v)

class test_expression_writer_ascii():

    expr = Expression([Symbol('eggs'), Symbol('ham')])
    repr = urepr = '(eggs ham)'

    def test_stringio_7(self):
        fp = StringIO()
        self.expr.print_into(fp)
        assert_equal(fp.getvalue(), self.repr)

    def test_stringio_8(self):
        fp = StringIO()
        self.expr.print_into(fp, escape_unicode=False)
        assert_equal(fp.getvalue(), self.urepr)

    def test_bytesio_7(self):
        fp = io.BytesIO()
        self.expr.print_into(fp)
        assert_equal(fp.getvalue(), b(self.repr))

    def test_bytesio_8(self):
        fp = io.BytesIO()
        self.expr.print_into(fp, escape_unicode=False)
        assert_equal(fp.getvalue(), b(self.urepr))

    def test_file_io_text_7(self):
        with tempfile.TemporaryFile(mode='w+t') as fp:
            self.expr.print_into(fp)
            fp.seek(0)
            assert_equal(fp.read(), self.repr)

    def test_file_io_text_8(self):
        if py3k:
            fp = tempfile.TemporaryFile(mode='w+t', encoding='UTF-16-LE')
        else:
            fp = tempfile.TemporaryFile(mode='w+t')
        with fp:
            self.expr.print_into(fp, escape_unicode=False)
            fp.seek(0)
            assert_equal(fp.read(), self.urepr)

    def test_codecs_io_text_7(self):
        tmpdir = tempfile.mkdtemp()
        try:
            path = os.path.join(tmpdir, 'tmp')
            with codecs.open(path, mode='w+', encoding='UTF-16-LE') as fp:
                self.expr.print_into(fp)
                fp.seek(0)
                assert_equal(fp.read(), self.repr)
        finally:
            shutil.rmtree(tmpdir)

    def test_codecs_io_text_8(self):
        tmpdir = tempfile.mkdtemp()
        try:
            path = os.path.join(tmpdir, 'tmp')
            with codecs.open(path, mode='w+', encoding='UTF-16-LE') as fp:
                self.expr.print_into(fp, escape_unicode=False)
                fp.seek(0)
                assert_equal(fp.read(), u(self.urepr))
        finally:
            shutil.rmtree(tmpdir)

    def test_file_io_binary_7(self):
        with tempfile.TemporaryFile(mode='w+b') as fp:
            self.expr.print_into(fp)
            fp.seek(0)
            assert_equal(fp.read(), b(self.repr))

    def test_file_io_binary_8(self):
        with tempfile.TemporaryFile(mode='w+b') as fp:
            self.expr.print_into(fp, escape_unicode=False)
            fp.seek(0)
            assert_equal(fp.read(), b(self.urepr))

    def test_as_string_7(self):
        s = self.expr.as_string()
        assert_equal(s, self.repr)

    def test_as_string_8(self):
        s = self.expr.as_string(escape_unicode=False)
        assert_equal(s, self.urepr)

class test_expression_writer_nonascii(test_expression_writer_ascii):

    expr = Expression(u('żółw'))
    repr = r'"\305\274\303\263\305\202w"'
    urepr = r'"żółw"'

def test_version():
    assert_is_instance(__version__, str)

def test_wildcard_import():
    ns = wildcard_import('djvu.sexpr')
    assert_list_equal(
        sorted(ns.keys()), [
            'Expression',
            'ExpressionSyntaxError',
            'IntExpression',
            'InvalidExpression',
            'ListExpression',
            'StringExpression',
            'Symbol',
            'SymbolExpression'
        ]
    )

# vim:ts=4 sts=4 sw=4 et
